# Copyright 2026 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Mapping, Sequence
import contextlib
import dataclasses
import functools
import itertools
import threading
import types
from typing import Any, Self

import jax
from jax import numpy as jnp
from jax._src import callback
from jax._src import source_info_util
from jax._src.pallas.mosaic.interpret import utils as interpret_utils
from jax._src.pallas.mosaic.interpret import vector_clock as vc
from jax._src.pallas.mosaic.interpret.race_detection_state import RaceDetectionState
from jax._src.pallas.mosaic_gpu import core as mosaic_gpu_core
from jax._src.pallas.mosaic_gpu.interpret import shared_memory as memory
from jax._src.pallas.mosaic_gpu.interpret.params import InterpretGPUParams
from jax._src.state import indexing
import numpy as np


IDX_BY_GPU_MEMORY_SPACE: Mapping[mosaic_gpu_core.MemorySpace, int]
IDX_BY_GPU_MEMORY_SPACE = types.MappingProxyType(
    {v: i for i, v in enumerate(mosaic_gpu_core.MemorySpace)}
)


GPU_MEMORY_SPACE_BY_IDX: Mapping[int, mosaic_gpu_core.MemorySpace]
GPU_MEMORY_SPACE_BY_IDX = types.MappingProxyType(
    dict(enumerate(mosaic_gpu_core.MemorySpace))
)


def get_memory_space_idx(space: mosaic_gpu_core.MemorySpace | None) -> int:
  if space is None:
    return IDX_BY_GPU_MEMORY_SPACE[mosaic_gpu_core.MemorySpace.SMEM]
  return IDX_BY_GPU_MEMORY_SPACE[space]


def is_gmem_memory_space(space: mosaic_gpu_core.MemorySpace | None) -> bool:
  return space == mosaic_gpu_core.MemorySpace.GMEM


_shared_memory: memory.GPUSharedMemory | None = None
_shared_memory_init_lock = threading.Lock()
_races: RaceDetectionState | None = None


def _get_shared_memory() -> memory.GPUSharedMemory:
  assert _shared_memory is not None
  return _shared_memory


def _clear_shared_memory():
  global _shared_memory
  with _shared_memory_init_lock:
    _shared_memory = None


def get_races() -> RaceDetectionState:
  assert _races is not None
  return _races


def reset_gpu_interpret_mode_state():
  """Resets all global, shared state used by GPU interpret mode.

  GPU interpret mode uses global, shared state for simulating memory buffers,
  for race detection, etc., when interpreting a kernel. Normally, this shared
  state is cleaned up after a kernel is interpreted.

  But if an exception is thrown while interpreting a kernel, the shared state
  is not cleaned up, allowing the simulated GPU state to be examined for
  debugging purposes. In this case, the shared state must be reset before
  any further kernels are interpreted.
  """
  global _shared_memory, _races
  with _shared_memory_init_lock:
    _shared_memory = None
    _races = None


# Below we define pairs of _callback_ functions. Each pair consists of
#
#   (1) a module-private function, e.g. `_initialize_shared_memory`, and
#   (2) a thin wrapper around the this module-private function, e.g.
#       `call_initialize_shared_memory`.
#
# The module-private function (1) runs in the Python ("host") process and
# manages interaction of the interpreted Pallas kernel with the memory system,
# represented by the module-global `SharedMemory` object `_shared_memory`.
#
# The wrapper function (2) is to be called from the interpreted Pallas kernel
# (that is simulating a "device", or thread). It serves as the interface between
# the "device" kernel and the "host" memory system and merely passes arguments
# on to the corresponding function (1). Importantly, when the wrapper receives
# an argument that is a Jax (device) array, this argument is received as a Numpy
# (host) array by the corresponding function (1), due to the
# `callback.io_callback` mechanism.


def _initialize_shared_memory(
    *,
    num_gpus: int,
    num_threads_per_block: int,
    num_blocks_per_cluster: int,
    interpret_params: InterpretGPUParams,
):
  global _shared_memory, _races

  num_gpus = int(num_gpus)
  num_threads_per_block = int(num_threads_per_block)
  num_total_concurrent_threads = (
      num_gpus * num_threads_per_block * num_blocks_per_cluster
  )

  with _shared_memory_init_lock:
    if _shared_memory is None:
      _races = RaceDetectionState(num_cores=num_total_concurrent_threads)
      _shared_memory = memory.GPUSharedMemory(
          num_devices=num_gpus,
          num_threads_per_block=num_threads_per_block,
          num_blocks_per_cluster=num_blocks_per_cluster,
          num_tma_threads_per_device=interpret_params.num_tma_threads_per_device,
          out_of_bounds_reads=interpret_params.out_of_bounds_reads,
          # TODO(nrink): Support different DMA execution modes on GPU.
          dma_execution_mode="eager",
          uninitialized_memory=interpret_params.uninitialized_memory,
          detect_races=interpret_params.detect_races,
          barrier=threading.Barrier(num_gpus, action=lambda: None),
          clean_up_barrier=threading.Barrier(
              num_gpus, action=_clear_shared_memory
          ),
          logging_mode=interpret_params.logging_mode,
      )
  # The naming of the `num_cores` property of `SharedMemory` originates from the
  # support for multipl cores in a (Megacore) TPU device. As commented above, on
  # GPU we model multiple Pallas threads per device as _cores_ in the
  # (TPU-/Megacore-)inspired terminology of `SharedMemory`.
  assert _shared_memory.num_cores == num_total_concurrent_threads


def call_initialize_shared_memory(
    *,
    num_gpus: int,
    num_threads_per_block: int,
    num_blocks_per_cluster: int,
    interpret_params: InterpretGPUParams,
):
  callback.io_callback(
      functools.partial(
          _initialize_shared_memory,
          interpret_params=interpret_params,
      ),
      (),
      num_gpus=num_gpus,
      num_threads_per_block=num_threads_per_block,
      num_blocks_per_cluster=num_blocks_per_cluster,
      ordered=True,
  )


def _clean_up_shared_memory():
  shared_memory = _get_shared_memory()
  shared_memory.clean_up_barrier.wait()


def call_clean_up_shared_memory():
  callback.io_callback(_clean_up_shared_memory, (), ordered=True)


def _update_clocks_for_device_barrier(device_id: int):
  shared_memory = _get_shared_memory()
  shared_memory.update_clocks_for_device_barrier(device_id)


def call_update_clocks_for_device_barrier(device_id: int):
  callback.io_callback(
      _update_clocks_for_device_barrier, (), device_id, ordered=True
  )


@dataclasses.dataclass(frozen=True, kw_only=True)
class HostAllocationRequest:
  """Request for an allocation on a device/thread and in a memory space."""

  memory_space_id: int
  device_id: int
  # Defaults to zero for `AllocationRequest`s that do not specify a thread ID.
  thread_id: int = 0
  # The reference count is needed only for allocations that are explicitly
  # deallocated (with _deallocate_buffer below). This currently only applies to
  # allocations made by a `run_scoped` primitive.
  initial_ref_count: int = 1

  def __iter__(self):
    # We make `self` iterable to ease conversion into Numpy and Jax arrays (cf.
    # methods `as_array` and `as_jax_array` below). Note that for this purpose
    # it would suffice to have any method that return a suitable iterator,
    # instead of implementing the special `__iter__` method. Not implementing
    # `__iter__` would mean that objects of this class cannot (accidentally) be
    # iterated over by clients of the class.
    return iter((
        self.memory_space_id,
        self.device_id,
        self.thread_id,
        self.initial_ref_count,
    ))

  @classmethod
  def shape_and_dtype(cls) -> jax.ShapeDtypeStruct:
    num_fields = len(dataclasses.fields(cls))
    return jax.ShapeDtypeStruct((num_fields,), jnp.int32)

  @property
  def as_array(self) -> np.ndarray:
    return np.array(list(self), dtype=np.int32)

  @property
  def as_jax_array(self) -> jnp.ndarray:
    return jnp.array(list(self), dtype=jnp.int32)

  @classmethod
  def from_array(cls, request: np.ndarray | jnp.ndarray) -> Self:
    if request.shape != cls.shape_and_dtype().shape:
      raise ValueError(
          f"Expected shape {cls.shape_and_dtype().shape} but got"
          f" {request.shape}"
      )
    if not interpret_utils.is_int(request.dtype):
      raise ValueError(f"Expected integer dtype but got {request.dtype}")

    arg_names = [f.name for f in dataclasses.fields(cls)]
    values = map(int, request)
    return cls(**dict(zip(arg_names, values)))


def make_allocation_request_array(
    *,
    memory_space_id: int,
    device_id: int,
    thread_id: int = 0,
    initial_ref_count: int = 1,
) -> jnp.ndarray:
  return HostAllocationRequest(
      memory_space_id=memory_space_id,
      device_id=device_id,
      thread_id=thread_id,
      initial_ref_count=initial_ref_count,
  ).as_jax_array


@dataclasses.dataclass(frozen=True, kw_only=True)
class HostAllocationKey(HostAllocationRequest):
  """Key for an allocation in shared memory."""

  buffer_id: int

  def __iter__(self):
    # Note that implementing `__iter__` here affects the bahviour of the
    # `as_array` and `as_jax_array` methods of the base class. This is intended.
    yield from super().__iter__()
    yield self.buffer_id


def _allocate_buffer_for_all_threads(
    device_id: Any,
    allocation_request: Any,
    value: Any,
    source_info: source_info_util.SourceInfo | None = None,
) -> np.ndarray:
  """Allocates a buffer for the given `allocation_request`.

  While only a single buffer is allocated, we increment the next buffer ID on
  `_shared_memory` for all threads. (This is analogous to the behavior when
  interpreting TPU kernels with multiple cores per TPU device.)

  Args:
    allocation_request: Array that converts into an `HostAllocationRequest` with
      `thread_id` set to zero. This requirement can be thought of as associating
      the allocated buffer (that is shared across all threads) with the zeroth
      thread.
    value: Array of values to initialize the allocated buffer with.

  Returns:
    `AllocationKey` to refer to the allocated buffer.

  Raises:
    ValueError: If the `thread_id` in `allocation_request` is not zero.
  """
  device_id = int(device_id)
  allocation_request = HostAllocationRequest.from_array(allocation_request)
  if allocation_request.thread_id != 0:
    raise ValueError(
        "`thread_id` must be zero when allocating a buffer for all threads"
    )
  value = np.array(value)
  shared_memory = _get_shared_memory()

  key: HostAllocationKey | None = None
  buffer_id: int | None = None
  for thread_id in range(shared_memory.num_cores_per_device):
    buffer_id_for_thread_id = shared_memory.get_next_buffer_id(
        device_id, thread_id
    )
    if not buffer_id:
      buffer_id = buffer_id_for_thread_id
    else:
      # We keep the buffer ids in sync across all threads. This implies, in
      # particular, that every instance of the assignment to `key` below assigns
      # an `AllocationKey` object with the same attributes.
      assert buffer_id == buffer_id_for_thread_id

    key = HostAllocationKey(
        memory_space_id=allocation_request.memory_space_id,
        device_id=allocation_request.device_id,
        thread_id=0,
        initial_ref_count=allocation_request.initial_ref_count,
        buffer_id=buffer_id,
    )
    ref_count = allocation_request.initial_ref_count
    # We rely on the fact that `allocate_buffer` will not allocate a new buffer
    # if one with the same key already exists.
    shared_memory.allocate_buffer(
        key,
        ref_count=ref_count,
        value=value,
        logging_info=interpret_utils.GPULoggingInfo(
            device_id=device_id,
            pallas_thread_id=0,
            source_info=source_info,
        ),
    )

  # We expect the `for`-loop above to have executed its body at least once.
  assert key is not None
  return key.as_array


def call_allocate_buffer_for_all_threads(
    device_id: int,
    allocation_request: jnp.ndarray,
    value: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
) -> jnp.ndarray:
  return callback.io_callback(
      functools.partial(
          _allocate_buffer_for_all_threads, source_info=source_info
      ),
      HostAllocationKey.shape_and_dtype(),
      device_id,
      allocation_request,
      value,
      ordered=True,
  )


def _allocate_buffer(
    device_id: Any,
    thread_id: Any,
    allocation_request: Any,
    value: Any,
    source_info: source_info_util.SourceInfo | None = None,
) -> np.ndarray:
  """Allocates a buffer for the given `allocation_request`.

  Args:
    allocation_request: Array that converts into a `HostAllocationRequest`.
    value: Array of values to initialize the allocated buffer with.
    source_info: Information about the source code location of the allocation.

  Returns:
    `AllocationKey` to refer to the allocated buffer.
  """
  device_id = int(device_id)
  thread_id = int(thread_id)
  allocation_request = HostAllocationRequest.from_array(allocation_request)
  value = np.array(value)
  shared_memory = _get_shared_memory()

  buffer_id = shared_memory.get_next_buffer_id(device_id, thread_id)

  key = HostAllocationKey(
      memory_space_id=allocation_request.memory_space_id,
      device_id=allocation_request.device_id,
      thread_id=allocation_request.thread_id,
      initial_ref_count=allocation_request.initial_ref_count,
      buffer_id=buffer_id,
  )
  ref_count = allocation_request.initial_ref_count
  shared_memory.allocate_buffer(
      key,
      ref_count=ref_count,
      value=value,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )
  return key.as_array


def call_allocate_buffer(
    device_id: int,
    thread_id: int,
    allocation_request: jnp.ndarray,
    value: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
) -> jnp.ndarray:
  return callback.io_callback(
      functools.partial(_allocate_buffer, source_info=source_info),
      HostAllocationKey.shape_and_dtype(),
      device_id,
      thread_id,
      allocation_request,
      value,
      ordered=True,
  )


def _deallocate_buffer(
    device_id: Any,
    thread_id: Any,
    allocation_key: Any,
    source_info: source_info_util.SourceInfo | None = None,
):
  """Decreases the reference count of the buffer with `allocation_key` (Deallocates the buffer if its reference count becomes zero)."""
  allocation_key = HostAllocationKey.from_array(allocation_key)
  shared_memory = _get_shared_memory()
  shared_memory.deallocate_buffer(
      allocation_key,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )


def call_deallocate_buffer(
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  callback.io_callback(
      functools.partial(_deallocate_buffer, source_info=source_info),
      None,
      device_id,
      thread_id,
      allocation_key,
      ordered=True,
  )


def _handle_out_of_bounds_read(
    ret: np.ndarray | None,
    full_read_shape: tuple[int, ...],
    shape: Sequence[int],
    dtype: np.dtype,
    allocation_key: HostAllocationKey,
    read_range: tuple[int | slice, ...],
    shared_memory: memory.GPUSharedMemory,
    source_info,
    input_name: str | None,
    block_indices: tuple[int, ...] | None,
    grid_loop_idx: tuple[int, ...] | None,
) -> np.ndarray:
  """Handles out-of-bounds read based on shared_memory configuration."""
  if shared_memory.out_of_bounds_reads == "raise":
    if source_info is None:
      ctx = contextlib.nullcontext()
    else:
      ctx = source_info_util.user_context(
          traceback=source_info.traceback, name_stack=source_info.name_stack
      )
    with ctx:
      if input_name is None:
        raise IndexError(
            f"Out-of-bounds read of {allocation_key}:"
            f" reading [{read_range}] but buffer has shape {shape}."
        )
      else:
        # Different error message when we are reading a block of an input,
        # to copy it to a buffer before invoking the kernel body.
        raise IndexError(
            f"Out-of-bounds block index {block_indices} for {allocation_key},"
            f' input "{input_name}" in iteration {grid_loop_idx}:'
            f" reading [{read_range}] but input has shape {shape}."
        )
  # out_of_bounds_reads == "uninitialized"
  uninit_array = np.full(
      full_read_shape,
      interpret_utils.get_uninitialized_value(
          dtype, shared_memory.uninitialized_memory
      ),
      dtype=dtype,
  )
  if ret is None:
    return uninit_array
  else:
    uninit_array[tuple(slice(s) for s in ret.shape)] = ret
    return uninit_array


def _is_dynamic(indexer: indexing.NDIndexer) -> bool:
  return any(
      isinstance(idx, indexing.Slice)
      and (idx.is_dynamic_start or idx.is_dynamic_size)
      for idx in indexer.indices
  )


def _validate_transforms(transforms):
  for transform in transforms:
    match transform:
      case indexing.NDIndexer():
        if _is_dynamic(transform):
          raise ValueError(
              "Dynamic indexing not supported in GPU interpret mode"
          )
      case _:
        raise ValueError(f"Unsupported transform: {transform}")


def _get(
    device_id: Any,
    thread_id: Any,
    allocation_key: Any,
    transforms,
    block_indices=None,
    grid_loop_idx=None,
    clock=None,
    increment_clock: bool = True,
    source_info=None,
    input_name=None,
) -> np.ndarray:
  """Performs a read from the buffer for `allocation_key_as_array` from the given device and thread."""
  device_id = int(device_id)
  thread_id = int(thread_id)
  allocation_key = HostAllocationKey.from_array(allocation_key)

  _validate_transforms(transforms)
  # TODO(nrink): Support tiling and swizzling transforms.
  transforms = jax.tree.map(int, transforms)

  if input_name is not None:
    # NOTE: input_name, block_indices, and grid_loop_idx are set only if this
    # function is being called to read a block from a pallas_call input (at the
    # start of one iteration of the kernel body).
    assert block_indices is not None
    block_indices = tuple(int(x) for x in block_indices)
    assert grid_loop_idx is not None
    grid_loop_idx = tuple(int(x) for x in grid_loop_idx)

  shared_memory = _get_shared_memory()

  global_core_id = shared_memory.get_global_core_id(device_id, thread_id)

  read_range = interpret_utils.to_range(transforms)
  ret, (shape, dtype), clock_ = shared_memory.get_buffer_content(
      allocation_key,
      read_range,
      global_core_id,
      increment_clock=increment_clock,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )
  clock = clock if clock is not None else clock_

  # Compute the shape of the read value, assuming the read is fully in-bounds.
  # TODO(jburnim): We already know this shape in the Jaxpr where we insert a
  # callback to `get`.  Should we just pass the shape to `get`?
  # TODO(jburnim): Move to a helper function?
  new_full_read_shape: list[int] = []
  assert len(read_range) <= len(shape)
  for dim_size, idx_or_slice in itertools.zip_longest(
      shape, read_range, fillvalue=None
  ):
    assert isinstance(dim_size, int)
    if idx_or_slice is None:
      new_full_read_shape.append(dim_size)
    elif isinstance(idx_or_slice, int):
      continue
    else:
      dim_size = (idx_or_slice.stop - idx_or_slice.start) // idx_or_slice.step
      assert isinstance(dim_size, int)
      new_full_read_shape.append(dim_size)
  full_read_shape = tuple(new_full_read_shape)
  del new_full_read_shape

  if (ret is None) or (full_read_shape != ret.shape):
    ret = _handle_out_of_bounds_read(
        ret,
        full_read_shape,
        shape,
        dtype,
        allocation_key,
        read_range,
        shared_memory,
        source_info,
        input_name,
        block_indices,
        grid_loop_idx,
    )

  if shared_memory.detect_races:
    get_races().check_read(
        device_id,
        thread_id,
        clock,
        allocation_key,
        read_range,
        source_info=source_info,
    )
  return ret


def call_get(
    *,
    result_shape_and_dtype,
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    transforms,
    block_indices=None,
    grid_loop_idx=None,
    clock=None,
    source_info=None,
    input_name=None,
) -> jnp.ndarray:
  return callback.io_callback(
      functools.partial(_get, source_info=source_info, input_name=input_name),
      result_shape_and_dtype,
      device_id,
      thread_id,
      allocation_key,
      transforms,
      block_indices,
      grid_loop_idx,
      clock,
      ordered=True,
  )


def _swap(
    device_id: Any,
    thread_id: Any,
    allocation_key_as_array: Any,
    transforms,
    val,
    mask,
    *,
    clock=None,
    increment_clock: bool = True,
    source_info=None,
):
  """Performs a swap into the buffer for `allocation_key_as_array` from the given device and thread."""
  device_id = int(device_id)
  thread_id = int(thread_id)
  allocation_key = HostAllocationKey.from_array(allocation_key_as_array)

  _validate_transforms(transforms)
  # TODO(nrink): Support tiling and swizzling transforms.
  transforms = jax.tree.map(int, transforms)

  val = np.array(val)
  mask = np.array(mask) if mask is not None else None
  if mask is not None:
    assert mask.shape == val.shape

  shared_memory = _get_shared_memory()

  global_core_id = shared_memory.get_global_core_id(device_id, thread_id)

  read_write_range = interpret_utils.to_range(transforms)
  ret, (shape, _), clock_ = shared_memory.swap_buffer_content(
      allocation_key,
      read_write_range,
      val,
      mask,
      global_core_id,
      increment_clock=increment_clock,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )
  clock = clock if clock is not None else clock_

  if ret is None:
    if mask is None:
      raise ValueError(
          f"Out-of-bounds swap of {allocation_key}:"
          f" swapping [{read_write_range}] but buffer has shape"
          f" {shape} ."
      )
    else:
      # TODO(jburnim): Include indices of out-of-bounds locations where mask
      # is True.
      raise ValueError(
          f"Out-of-bounds masked swap of {allocation_key}: swapping"
          f" [{read_write_range}] but buffer has shape {shape} . "
      )

  if shared_memory.detect_races:
    get_races().check_write(
        device_id,
        thread_id,
        clock,
        allocation_key,
        read_write_range,
        source_info=source_info,
    )
  return ret


def call_swap(
    *,
    result_shape_and_dtype,
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    transforms,
    val,
    mask,
    clock=None,
    source_info=None,
):
  return callback.io_callback(
      functools.partial(_swap, source_info=source_info),
      result_shape_and_dtype,
      device_id,
      thread_id,
      allocation_key,
      transforms,
      val,
      mask,
      clock=clock,
      ordered=True,
  )


def _allocate_barriers(
    device_id: Any,
    thread_id: Any,
    num_arrivals: Any,
    num_barriers: Any,
    ref_count: Any,
    source_info: source_info_util.SourceInfo | None = None,
) -> np.ndarray:
  device_id = int(device_id)
  thread_id = int(thread_id)
  num_arrivals = int(num_arrivals)
  num_barriers = int(num_barriers)
  ref_count = int(ref_count)
  shared_memory = _get_shared_memory()

  keys = []
  for _ in range(num_barriers):
    # Advance `shared_memory`'s internal buffer id counter for all threads that
    # call into this function.
    barrier_id = shared_memory.get_next_buffer_id(device_id, thread_id)
    smem_space_id = IDX_BY_GPU_MEMORY_SPACE[mosaic_gpu_core.SMEM]
    key = HostAllocationKey(
        memory_space_id=smem_space_id,
        device_id=device_id,
        # Barriers are shared between threads. Hence we associate all
        # allocations for `Barrier`s with the 0th thread.
        thread_id=0,
        initial_ref_count=ref_count,
        buffer_id=barrier_id,
    )

    shared_memory.allocate_barrier(
        key,
        ref_count=ref_count,
        num_arrivals=num_arrivals,
        logging_info=interpret_utils.GPULoggingInfo(
            device_id=device_id,
            pallas_thread_id=thread_id,
            source_info=source_info,
        ),
    )
    keys.append(key.as_array)

  assert len(keys) == num_barriers
  return np.array(keys, dtype=np.int32)


def call_allocate_barriers(
    device_id: int,
    thread_id: int,
    num_arrivals: int,
    num_barriers: int,
    ref_count: int,
    source_info: source_info_util.SourceInfo | None = None,
) -> jnp.ndarray:
  shape_and_dtype = HostAllocationKey.shape_and_dtype()
  result_shape = (num_barriers, *shape_and_dtype.shape)
  result_shape_and_dtype = jax.ShapeDtypeStruct(
      result_shape, shape_and_dtype.dtype
  )
  return callback.io_callback(
      functools.partial(_allocate_barriers, source_info=source_info),
      result_shape_and_dtype,
      device_id,
      thread_id,
      num_arrivals,
      num_barriers,
      ref_count,
      ordered=True,
  )


def _deallocate_barrier(
    device_id: Any,
    thread_id: Any,
    allocation_key: np.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  device_id = int(device_id)
  thread_id = int(thread_id)

  assert len(allocation_key.shape) == 2
  num_barriers = allocation_key.shape[0]

  keys_to_deallocate = []
  for i in range(num_barriers):
    keys_to_deallocate.append(allocation_key[i, :])

  shared_memory = _get_shared_memory()

  for key in keys_to_deallocate:
    barrier_allocation_key = HostAllocationKey.from_array(key)
    shared_memory.deallocate_barrier(
        barrier_allocation_key,
        logging_info=interpret_utils.GPULoggingInfo(
            device_id=device_id,
            pallas_thread_id=thread_id,
            source_info=source_info,
        ),
    )


def call_deallocate_barrier(
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  callback.io_callback(
      functools.partial(_deallocate_barrier, source_info=source_info),
      None,
      device_id,
      thread_id,
      allocation_key,
      ordered=True,
  )


def _barrier_wait(
    device_id: int,
    thread_id: int,
    allocation_key: np.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  device_id = int(device_id)
  thread_id = int(thread_id)
  barrier_key = HostAllocationKey.from_array(allocation_key)
  shared_memory = _get_shared_memory()

  barrier, _ = shared_memory.get_barrier_and_increment_clock(
      barrier_key, device_id, thread_id
  )
  barrier.wait(
      device_id,
      thread_id,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )


def call_barrier_wait(
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  callback.io_callback(
      functools.partial(_barrier_wait, source_info=source_info),
      None,
      device_id,
      thread_id,
      allocation_key,
      ordered=True,
  )


def _barrier_arrive(
    device_id: int,
    thread_id: int,
    allocation_key: np.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  device_id = int(device_id)
  thread_id = int(thread_id)
  barrier_key = HostAllocationKey.from_array(allocation_key)
  shared_memory = _get_shared_memory()

  barrier, clock = shared_memory.get_barrier_and_increment_clock(
      barrier_key, device_id, thread_id
  )
  barrier.arrive(
      clock,
      logging_info=interpret_utils.GPULoggingInfo(
          device_id=device_id,
          pallas_thread_id=thread_id,
          source_info=source_info,
      ),
  )


def call_barrier_arrive(
    device_id: int,
    thread_id: int,
    allocation_key: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  callback.io_callback(
      functools.partial(_barrier_arrive, source_info=source_info),
      None,
      device_id,
      thread_id,
      allocation_key,
      ordered=True,
  )


def _assert_no_barriers_allocated():
  _get_shared_memory().assert_no_barriers_allocated()


def call_assert_no_barriers_allocated():
  callback.io_callback(_assert_no_barriers_allocated, (), ordered=True)


@dataclasses.dataclass(kw_only=True)
class DeviceLocalMemoryTransfer:
  """Represents a device-local memory transfer, e.g. GMEM to SMEM."""

  # The device ID of the device on which the memory transfer is being executed.
  device_id: int
  # The thread ID of the thread that has initiated the memory transfer.
  thread_id: int

  # Allocation key (and transforms) for the source buffer.
  src_allocation_key: np.ndarray
  src_transforms: tuple[Any, ...]

  # Allocation key (and transforms) for the destination buffer.
  dst_allocation_key: np.ndarray
  dst_transforms: tuple[Any, ...]

  # Allocation key for the barrier that should be used to synchronize the
  # execution of the memory transfer.
  barrier_allocation_key_as_array: np.ndarray

  source_info: source_info_util.SourceInfo | None = None

  # `data` is only used to track internal state of the `MemoryTransfer`
  # object. Hence `data` must not be explicitly initialized, as checked in
  # `__post_init__` below.
  data: np.ndarray | None = None

  clock: vc.VectorClock | None = None

  def __post_init__(self):
    assert self.data is None

  def execute(self):
    """Executes the memory transfer (both reading and writing parts).

    Note that the caller must not hold the lock on the shared memory (because
    `get` and `swap` are called in this method).
    """

    shared_memory = _get_shared_memory()

    barrier_key = HostAllocationKey.from_array(
        self.barrier_allocation_key_as_array
    )
    barrier, clock = shared_memory.get_barrier_and_increment_clock(
        barrier_key, self.device_id, self.thread_id
    )

    tma_thread_id = shared_memory.get_next_tma_thread_id(self.device_id)
    global_tma_thread_id = shared_memory.get_global_thread_id(
        self.device_id, tma_thread_id
    )
    # We use the `clock` from the barrier lookup to synchronize the TMA thread
    # with the Pallas thread that initiated the transfer, i.e. the thread with
    # `self.thread_id`. Note that the barrier lookup was done with
    # `self.thread_id`, and hence `clock` is the vector clock for the thread
    # with `self.thread_id`.
    if shared_memory.detect_races:
      self.clock = clock

    # A memory transfer should be executed only once. We check this by asserting
    # that `self.data` is `None` at the start of the transfer.
    assert self.data is None

    self.data = _get(
        device_id=np.array(self.device_id),
        thread_id=np.array(self.thread_id),
        clock=vc.copy_vector_clock(self.clock),
        increment_clock=False,
        allocation_key=self.src_allocation_key,
        transforms=self.src_transforms,
        source_info=self.source_info,
    )
    assert self.data is not None
    if shared_memory.detect_races:
      assert self.clock is not None
      vc.inc_vector_clock(self.clock, global_tma_thread_id)

    # We write `self.data` to the destination allocation using `_swap`, where
    # the result (i.e. the old contents of the destination buffer) is ignored.
    _swap(
        device_id=np.array(self.device_id),
        thread_id=np.array(self.thread_id),
        clock=vc.copy_vector_clock(self.clock),
        increment_clock=False,
        allocation_key_as_array=self.dst_allocation_key,
        transforms=self.dst_transforms,
        val=self.data,
        mask=None,
        source_info=self.source_info,
    )
    if shared_memory.detect_races:
      assert self.clock is not None
      vc.inc_vector_clock(self.clock, global_tma_thread_id)

    barrier.arrive(
        clock=vc.copy_vector_clock(self.clock),
        logging_info=interpret_utils.GPULoggingInfo(
            device_id=self.device_id,
            pallas_thread_id=tma_thread_id,
            source_info=self.source_info,
        ),
    )


# TODO(nrink): Once non-eager execution of memory transfers/DMAs is supported in
# GPU interpret mode, consider renaming this function to something along the
# lines of `_enqueue_device_local_memory_transfer`.
def _execute_device_local_memory_transfer(
    *,
    device_id: np.ndarray,
    thread_id: np.ndarray,
    src_allocation_key: np.ndarray,
    src_transforms: tuple[Any, ...],
    dst_allocation_key: np.ndarray,
    dst_transforms: tuple[Any, ...],
    barrier_allocation_key_as_array: np.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  device_id_as_int = int(device_id)
  thread_id_as_int = int(thread_id)

  transfer = DeviceLocalMemoryTransfer(
      device_id=device_id_as_int,
      thread_id=thread_id_as_int,
      src_allocation_key=src_allocation_key,
      src_transforms=src_transforms,
      dst_allocation_key=dst_allocation_key,
      dst_transforms=dst_transforms,
      barrier_allocation_key_as_array=barrier_allocation_key_as_array,
      source_info=source_info,
  )
  transfer.execute()


# TODO(nrink): Once non-eager execution of memory transfers/DMAs is supported in
# GPU interpret mode, consider renaming this function to something along the
# lines of `call_enqueue_device_local_memory_transfer`.
def call_execute_device_local_memory_transfer(
    *,
    device_id: int,
    thread_id: int,
    src_allocation_key: jnp.ndarray,
    src_transforms: tuple[Any, ...],
    dst_allocation_key: jnp.ndarray,
    dst_transforms: tuple[Any, ...],
    barrier_allocation_key_as_array: jnp.ndarray,
    source_info: source_info_util.SourceInfo | None = None,
):
  callback.io_callback(
      functools.partial(
          _execute_device_local_memory_transfer, source_info=source_info
      ),
      None,
      device_id=device_id,
      thread_id=thread_id,
      src_allocation_key=src_allocation_key,
      src_transforms=src_transforms,
      dst_allocation_key=dst_allocation_key,
      dst_transforms=dst_transforms,
      barrier_allocation_key_as_array=barrier_allocation_key_as_array,
      ordered=True,
  )



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains TPU-specific Pallas abstractions."""

from __future__ import annotations

import collections
from collections.abc import Mapping, Sequence
import dataclasses
import enum
from typing import Any, Literal

import jax
from jax._src import core as jax_core
from jax._src import linear_util as lu
from jax._src import state
from jax._src import util
from jax._src.frozen_dict import FrozenDict
from jax._src.interpreters import partial_eval as pe
from jax._src.pallas import core as pallas_core
import jax.numpy as jnp
import numpy as np


map, unsafe_map = util.safe_map, map
zip, unsafe_zip = util.safe_zip, zip

no_block_spec = pallas_core.no_block_spec
_out_shape_to_aval_mapping = pallas_core._out_shape_to_aval_mapping


class CoreType(enum.Enum):
  TC = "tc"
  SC_SCALAR_SUBCORE = "sc_scalar_subcore"
  SC_VECTOR_SUBCORE = "sc_vector_subcore"

  def __str__(self) -> str:
    return self.value

  def __repr__(self) -> str:
    return self.name


class GridDimensionSemantics(enum.Enum):
  PARALLEL = "parallel"
  CORE_PARALLEL = "core_parallel"
  SUBCORE_PARALLEL = "subcore_parallel"
  ARBITRARY = "arbitrary"


PARALLEL = GridDimensionSemantics.PARALLEL
CORE_PARALLEL = GridDimensionSemantics.CORE_PARALLEL
SUBCORE_PARALLEL = GridDimensionSemantics.SUBCORE_PARALLEL
ARBITRARY = GridDimensionSemantics.ARBITRARY


LiteralDimensionSemantics = Literal[
    "parallel", "core_parallel", "subcore_parallel", "arbitrary"
]
DimensionSemantics = LiteralDimensionSemantics | GridDimensionSemantics


class SideEffectType(enum.Enum):
  # No side effects, can be deduplicated / removed if unused.
  PURE = "pure"
  # Cannot be deduplicated, but can be removed if unused.
  DATAFLOW_SIDE_EFFECTING = "dataflow_side_effecting"
  # Cannot be deduplicated or removed.
  SIDE_EFFECTING = "side_effecting"


@dataclasses.dataclass(frozen=True)
class CompilerParams:
  """Mosaic TPU compiler parameters.

  Attributes:
    dimension_semantics: A list of dimension semantics for each grid dimension
      of the kernel. Either "parallel" for dimensions that can execute in any
      order, or "arbitrary" for dimensions that must be executed sequentially.
    allow_input_fusion: A list of booleans indicating whether input fusion is
      allowed for each argument.
    vmem_limit_bytes: Overrides the default VMEM limit for a kernel. Note that
      this must be used in conjunction with the
      --xla_tpu_scoped_vmem_limit_kib=N flag with N*1kib > vmem_limit_bytes.
    collective_id: Indicates which barrier semaphore to use for the kernel. Note
      that using the same collective_id does not guarantee that the same barrier
      semaphore will be allocated between kernels.
    has_side_effects: Set to True to prevent kernel being CSEd by XLA.
    flags: A dictionary of command line flags for the kernel.
    internal_scratch_in_bytes: The size of the internal scratch space used by
      Mosaic.
    serialization_format: The serialization format for the kernel body.
    kernel_type: Specify if the kernel is meant to run on TensorCore or one of
      the SparseCores
    disable_bounds_checks: Disable bounds checks in the kernel.
    disable_semaphore_checks: Disable semaphore checks in the kernel.
    skip_device_barrier: Skip the default device barrier for the kernel.
    allow_collective_id_without_custom_barrier: Allow the use of collective_id
      without a custom barrier.
    use_tc_tiling_on_sc: Use TensorCore tiling for SparseCore. This flag is only
      used for ``SC_*_SUBCORE`` kernels.
  """

  dimension_semantics: tuple[DimensionSemantics, ...] | None = None
  allow_input_fusion: tuple[bool, ...] | None = None
  vmem_limit_bytes: int | None = None
  collective_id: int | None = None
  has_side_effects: bool | SideEffectType = False
  flags: dict[str, Any] | None = None
  internal_scratch_in_bytes: int | None = None
  serialization_format: int = 1
  kernel_type: CoreType = CoreType.TC
  disable_bounds_checks: bool = False
  disable_semaphore_checks: bool = False
  skip_device_barrier: bool = False
  allow_collective_id_without_custom_barrier: bool = False
  shape_invariant_numerics: bool = True
  use_tc_tiling_on_sc: bool | None = None

  def __init__(
      self,
      dimension_semantics: Sequence[DimensionSemantics] | None = None,
      allow_input_fusion: Sequence[bool] | None = None,
      vmem_limit_bytes: int | None = None,
      collective_id: int | None = None,
      has_side_effects: bool | SideEffectType = False,
      flags: Mapping[str, Any] | None = None,
      internal_scratch_in_bytes: int | None = None,
      serialization_format: int = 1,
      kernel_type: CoreType = CoreType.TC,
      disable_bounds_checks: bool = False,
      disable_semaphore_checks: bool = False,
      skip_device_barrier: bool = False,
      allow_collective_id_without_custom_barrier: bool = False,
      shape_invariant_numerics: bool = True,
      use_tc_tiling_on_sc: bool | None = None,
  ):
    object.__setattr__(
        self,
        "dimension_semantics",
        None if dimension_semantics is None else tuple(dimension_semantics),
    )
    object.__setattr__(
        self,
        "allow_input_fusion",
        None if allow_input_fusion is None else tuple(allow_input_fusion),
    )
    object.__setattr__(self, "vmem_limit_bytes", vmem_limit_bytes)
    object.__setattr__(self, "collective_id", collective_id)
    object.__setattr__(self, "has_side_effects", has_side_effects)
    object.__setattr__(
        self, "flags", None if flags is None else FrozenDict(flags)
    )
    object.__setattr__(
        self, "internal_scratch_in_bytes", internal_scratch_in_bytes
    )
    object.__setattr__(self, "serialization_format", serialization_format)
    object.__setattr__(self, "kernel_type", kernel_type)
    object.__setattr__(self, "disable_bounds_checks", disable_bounds_checks)
    object.__setattr__(
        self, "disable_semaphore_checks", disable_semaphore_checks
    )
    object.__setattr__(self, "skip_device_barrier", skip_device_barrier)
    object.__setattr__(
        self,
        "allow_collective_id_without_custom_barrier",
        allow_collective_id_without_custom_barrier,
    )
    object.__setattr__(
        self, "shape_invariant_numerics", shape_invariant_numerics
    )
    object.__setattr__(self, "use_tc_tiling_on_sc", use_tc_tiling_on_sc)

  # Replace is a method, not a field.
  replace = dataclasses.replace


class MemoryRef(pallas_core.MemoryRef):

  def __matmul__(self, other, /):
    if not isinstance(other, CoreType):
      return NotImplemented
    return dataclasses.replace(self, memory_space=self.memory_space @ other)


class MemorySpace(enum.Enum):
  VMEM = "vmem"
  VMEM_SHARED = "vmem_shared"
  SMEM = "smem"
  CMEM = "cmem"
  SEMAPHORE = "semaphore_mem"
  HBM = "hbm"
  HOST = "host"

  def __str__(self) -> str:
    return self.value

  def __repr__(self) -> str:
    return self.name

  def from_type(self, ty):
    return MemoryRef(ty, memory_space=self)

  def __call__(self, shape: Sequence[int], dtype: jnp.dtype[Any]):
    # A convenience function for constructing MemoryRef types of ShapedArrays.
    return self.from_type(jax_core.ShapedArray(tuple(shape), dtype))

  def __matmul__(self, other, /):
    if not isinstance(other, CoreType):
      return NotImplemented
    return CoreMemorySpace(self, other)


@dataclasses.dataclass(frozen=True)
class CoreMemorySpace:
  """A memory space tied to a specific core type."""

  memory_space: MemorySpace
  core_type: CoreType

  def __post_init__(self):
    match self.memory_space, self.core_type:
      case MemorySpace.VMEM, CoreType.TC | CoreType.SC_VECTOR_SUBCORE:
        ...
      case MemorySpace.SMEM | MemorySpace.SEMAPHORE, _:
        ...
      case MemorySpace.CMEM, CoreType.TC:
        ...
      case _, _:
        raise ValueError(
            "Unsupported core memory space:"
            f" {self.memory_space, self.core_type}"
        )

  def __call__(self, shape: Sequence[int], dtype: jnp.dtype[Any]):
    return MemoryRef(jax_core.ShapedArray(tuple(shape), dtype), self)

  def __str__(self) -> str:
    return f"{self.memory_space}@{self.core_type}"

  def __repr__(self) -> str:
    return f"{self.memory_space!r}@{self.core_type!r}"


class dma_semaphore(pallas_core.semaphore_dtype):
  pass


class DMASemaphore(pallas_core.AbstractSemaphoreTy):
  type = dma_semaphore
  name = "dma_sem"


class SemaphoreType(enum.Enum):
  REGULAR = "regular"
  DMA = "dma"
  BARRIER = "barrier"

  @property
  def dtype(self) -> Any:
    if self == SemaphoreType.DMA:
      return DMASemaphore()
    elif self == SemaphoreType.BARRIER:
      return pallas_core.BarrierSemaphore()
    else:
      return pallas_core.Semaphore()

  def __call__(self, shape: tuple[int, ...]):
    return MemoryRef(jax_core.ShapedArray(shape, self.dtype), MemorySpace.SEMAPHORE)

  def __matmul__(self, other, /):
    if not isinstance(other, CoreType):
      return NotImplemented
    return CoreMemorySpace(MemorySpace.SEMAPHORE, other)((), self.dtype)

  def get_array_aval(self) -> jax_core.ShapedArray:
    return self(()).get_array_aval()

  def get_ref_aval(self) -> state.AbstractRef:
    return self(()).get_ref_aval()


@dataclasses.dataclass(frozen=True)
class AbstractSemaphore(jax_core.AbstractValue):
  sem_type: SemaphoreType


@dataclasses.dataclass(init=False, kw_only=True, unsafe_hash=True)
class PrefetchScalarGridSpec(pallas_core.GridSpec):
  num_scalar_prefetch: int

  def __init__(
      self,
      num_scalar_prefetch: int,
      grid: pallas_core.Grid = (),
      in_specs: pallas_core.BlockSpecTree = no_block_spec,
      out_specs: pallas_core.BlockSpecTree = no_block_spec,
      scratch_shapes: pallas_core.ScratchShapeTree = (),
  ):
    super().__init__(grid, in_specs, out_specs, scratch_shapes)
    self.num_scalar_prefetch = num_scalar_prefetch
    self.scratch_shapes = tuple(scratch_shapes)

  def _make_scalar_ref_aval(self, aval):
    return state.AbstractRef(
        jax_core.ShapedArray(aval.shape, aval.dtype), MemorySpace.SMEM
    )


@dataclasses.dataclass(frozen=True)
class TensorCore:
  id: int


@dataclasses.dataclass(frozen=True)
class TensorCoreMesh(pallas_core.Mesh):
  """A mesh of TensorCores."""

  devices: np.ndarray
  axis_names: Sequence[str]

  def __init__(self, devices: np.ndarray, axis_names: Sequence[str]):
    devices = np.copy(devices)
    devices.setflags(write=False)
    object.__setattr__(self, "devices", devices)
    object.__setattr__(self, "axis_names", tuple(axis_names))

  def __hash__(self) -> int:
    return hash(
        (self.devices.shape, tuple(np.ravel(self.devices)), self.axis_names)
    )

  @property
  def kernel_type(self) -> CoreType:
    return CoreType.TC

  @property
  def default_memory_space(self) -> pallas_core.MemorySpace:
    return pallas_core.MemorySpace.ANY

  @property
  def shape(self):
    return collections.OrderedDict(zip(self.axis_names, self.devices.shape))

  @property
  def dimension_semantics(self) -> Sequence[str]:
    return ["parallel"]

  def discharges_effect(self, effect: jax_core.Effect) -> Literal[False]:
    del effect
    return False

  def check_is_compatible_with(self, other_mesh):
    if isinstance(other_mesh, TensorCoreMesh) and self != other_mesh:
      raise ValueError("You can't use two different TensorCoreMeshes.")
    # TODO: Add support for mpmd with SparseCore meshes.
    return super().check_is_compatible_with(other_mesh)

def create_tensorcore_mesh(
    axis_name: str,
    devices: Sequence[jax.Device] | None = None,
    num_cores: int | None = None,
) -> TensorCoreMesh:
  if devices is not None and num_cores is not None:
    raise ValueError("cannot specify both devices and num_cores")
  if num_cores is None:
    if devices is None:
      abstract_device = jax.sharding.get_abstract_mesh().abstract_device
      if abstract_device is None:
        devices = [jax.devices()[0]]
      else:
        devices = [abstract_device]
    num_cores = devices[0].num_cores
  return TensorCoreMesh(
      np.array([TensorCore(i) for i in range(num_cores)]),
      [axis_name],
  )


def pass_scalars_as_refs(
    jaxpr: jax_core.Jaxpr,
    args: Sequence[Any],
    in_avals: Sequence[jax_core.AbstractValue],
    out_avals: Sequence[jax_core.AbstractValue],
    mesh,
    copy_to_smem: bool = False,
) -> tuple[
    jax_core.Jaxpr,
    tuple[Any, ...],
    tuple[jax_core.AbstractValue, ...],
    tuple[jax_core.AbstractValue, ...],
    tuple[bool, ...],
]:
  """Rewrites a jaxpr to pass scalars as refs instead of values."""
  def allowed_aval(aval):
    if isinstance(aval, state.AbstractRef):
      return True
    if isinstance(aval, jax_core.ShapedArray):
      # Only scalars are allowed.
      return not aval.shape
    return False

  assert all(allowed_aval(v.aval) for v in jaxpr.constvars + jaxpr.invars)

  is_scalar_const = [
      isinstance(v.aval, jax_core.ShapedArray) and not v.aval.shape
      for v in jaxpr.constvars
  ]
  if not any(is_scalar_const):
    return (
        jaxpr,
        tuple(in_avals),
        tuple(out_avals),
        tuple(args),
        tuple(is_scalar_const),
    )
  non_scalar_const_avals, scalar_const_avals = util.partition_list(
      is_scalar_const,
      [v.aval for v in jaxpr.constvars],
  )
  non_scalar_consts, scalar_consts = util.partition_list(
      is_scalar_const, args
  )
  if copy_to_smem:
    smem_alloc = [
        state.AbstractRef(
            jax_core.ShapedArray((1,), aval.dtype),  # pyrefly: ignore[missing-attribute]
            memory_space=MemorySpace.SMEM,
        )
        for aval in scalar_const_avals
    ]
  else:
    smem_alloc = []

  # Rewrite body jaxpr to take in scalar values as Refs.
  def new_body(*args):
    scalar_const_refs, non_scalar_const_refs, args = util.split_list(
        args, [len(scalar_consts), len(non_scalar_consts)]
    )
    if copy_to_smem:
      smem, args = util.split_list(args, [len(smem_alloc)])
      assert len(smem) == len(scalar_const_refs)
      from jax._src.pallas.mosaic.helpers import sync_copy

      sync_copy(scalar_const_refs, smem)
    else:
      smem = scalar_const_refs
    scalar_const_values = [s[0] for s in smem]
    new_consts = util.merge_lists(
        is_scalar_const, non_scalar_const_refs, scalar_const_values
    )
    return jax_core.eval_jaxpr(jaxpr, new_consts, *args)

  # TODO(sharadmv): Remove this once Mosaic support passing scalars as values.
  scalar_const_trace_avals = [
      state.AbstractRef(
          jax_core.ShapedArray((1,), aval.dtype),  # pyrefly: ignore[missing-attribute]
          memory_space=MemorySpace.HBM if copy_to_smem else MemorySpace.SMEM,
      )
      for aval in scalar_const_avals
  ]
  new_trace_avals = [
      *scalar_const_trace_avals,
      *non_scalar_const_avals,
      *smem_alloc,
      *[v.aval for v in jaxpr.invars],
  ]
  with (
      pallas_core.tracing_grid_env(
          tuple(mesh.shape.values()), mapped_dims=()
      ),
      jax_core.extend_axis_env_nd(mesh.shape.items()),
  ):
    new_jaxpr, _, _ = pe.trace_to_jaxpr_dynamic(
        lu.wrap_init(
            new_body, debug_info=jaxpr.debug_info.with_unknown_names()
        ),
        new_trace_avals,
    )
  jaxpr = new_jaxpr.replace(
      constvars=new_jaxpr.invars[: len(jaxpr.constvars)],
      invars=new_jaxpr.invars[len(jaxpr.constvars) :],
  )
  args = [
      *[a[None] for a in scalar_consts],
      *non_scalar_consts,
  ]
  in_avals, out_avals, _ = util.split_list(
      new_trace_avals, [len(in_avals), len(out_avals)]
  )
  return jaxpr, tuple(in_avals), tuple(out_avals), tuple(args), tuple(is_scalar_const)


def _tensorcore_mesh_discharge_rule(
    in_avals,
    out_avals,
    *args,
    mesh,
    jaxpr,
    compiler_params: Any | None,
    interpret: Any,
    debug: bool,
    cost_estimate: pallas_core.CostEstimate | None,
    name: str,
    metadata: FrozenDict[str, str] | None,
):
  assert isinstance(mesh, TensorCoreMesh)
  if compiler_params and not isinstance(compiler_params, CompilerParams):
    raise ValueError("compiler_params must be a pltpu.CompilerParams")
  if not compiler_params:
    compiler_params = CompilerParams()
  if len(mesh.shape) > 1:
    raise NotImplementedError("Mesh must be 1D")
  if compiler_params.dimension_semantics is not None:
    raise ValueError("dimension_semantics must be None for TensorCoreMesh")
  num_cores = len(mesh.devices)
  if num_cores > 1:
    # Since each core will have its own VMEM, we currently disallow VMEM inputs
    # and outputs since other ops might not agree on how they are sharded across
    # cores by the (core-mapped) kernel.
    if any(
        pallas_core.get_memory_space_aval(aval) == MemorySpace.VMEM
        for aval in in_avals
    ):
      raise NotImplementedError(
          "TensorCoreMesh does not support VMEM inputs/outputs when there are"
          " >1 cores. Use HBM or ANY instead."
      )
  jaxpr, in_avals, out_avals, args, is_scalar_const = pass_scalars_as_refs(
      jaxpr, args, in_avals, out_avals, mesh
  )
  refs_out, out = pallas_core.default_mesh_discharge_rule(
      in_avals,
      out_avals,
      *args,
      jaxpr=jaxpr,
      mesh=mesh,
      compiler_params=compiler_params,
      debug=debug,
      interpret=interpret,
      cost_estimate=cost_estimate,
      name=name,
      metadata=metadata,
  )
  refs_out = [
      a if not is_scalar else None
      for is_scalar, a in zip(is_scalar_const, refs_out)
  ]
  return refs_out, out


pallas_core._core_map_mesh_rules[TensorCoreMesh] = (
    _tensorcore_mesh_discharge_rule
)


def _convert_semaphore_type_to_aval(
    out_shape: SemaphoreType,
) -> jax_core.AbstractValue:
  return out_shape.get_array_aval()


pallas_core._out_shape_to_aval_mapping[SemaphoreType] = (
    _convert_semaphore_type_to_aval
)


def memory_space_to_tpu_memory_space(
    memory_space: (
        MemorySpace | pallas_core.MemorySpace | CoreMemorySpace | None
    ),
    kernel_type: CoreType,
) -> MemorySpace | pallas_core.MemorySpace | CoreMemorySpace:
  match memory_space:
    case None:
      match kernel_type:
        case CoreType.TC | CoreType.SC_VECTOR_SUBCORE:
          return MemorySpace.VMEM
        case CoreType.SC_SCALAR_SUBCORE:
          return MemorySpace.SMEM
        case _:
          raise ValueError(f"Unsupported kernel type: {kernel_type}")
    case pallas_core.MemorySpace.ANY:
      return pallas_core.MemorySpace.ANY
    case pallas_core.MemorySpace.HOST:
      return MemorySpace.HOST
    case (
        pallas_core.MemorySpace.ERROR
        | pallas_core.MemorySpace.INDEX
        | pallas_core.MemorySpace.KEY
    ):
      return MemorySpace.SMEM
    case CoreMemorySpace():
      return (
          memory_space.memory_space
          if memory_space.core_type is kernel_type
          else memory_space
      )
    case MemorySpace():
      return memory_space
    case _:
      raise ValueError(f"Invalid memory space: {memory_space!r}")



# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/jax-ml/jax/issues/7570

from jax._src.ffi import (
    build_ffi_lowering_function as build_ffi_lowering_function,
    ffi_call as ffi_call,
    ffi_lowering as ffi_lowering,
    include_dir as include_dir,
    pycapsule as pycapsule,
    register_ffi_target as register_ffi_target,
    register_ffi_type_id as register_ffi_type_id,
    register_ffi_type as register_ffi_type,
    register_ffi_target_as_batch_partitionable as register_ffi_target_as_batch_partitionable,
)



# Copyright 2026 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import partial
from typing import Callable

from jax._src import core
from jax._src import api_util
from jax._src.util import safe_map, safe_zip, unzip2, weakref_lru_cache
from jax._src.interpreters import partial_eval as pe
from jax._src.tree_util import (
    FlatTree, Partial, tree_unflatten, tree_leaves_checked)
from jax._src import source_info_util
from jax._src.core import typeof

map = safe_map
zip = safe_zip

# TODO
#  [ ] static_argnums and static_argnames (via FlatTree)
#  [ ] allow NotAvailable sentinels
#  [ ] primal-output-to-residual forwarding

def remat_transform(policy, f, *args):
  dbg = api_util.debug_info("remat", f, args, {})
  with core.take_current_trace() as parent_trace:
    jaxpr_trace = pe.DynamicJaxprTrace(None)
    trace = RematTrace(parent_trace, jaxpr_trace, core.TraceTag(), policy)
    args_ft = FlatTree.flatten_static_argnums_argnames(args, {}, (), ())
    in_tracers = args_ft.map(
        # pyrefly: ignore[bad-argument-type]
        lambda x: RematTracer(trace, x, jaxpr_trace.new_arg(typeof(x), None)))  # noqa F821
    with core.set_current_trace(trace):
      args, kwargs = in_tracers.unflatten()
      ans_pytree = f(*args, **kwargs)
      dbg = dbg.set_result_paths(ans_pytree)
      ans_ft = FlatTree.flatten(ans_pytree)
      del ans_pytree, args, kwargs
    out_ft, out_tracer_ft = ans_ft.map(trace.to_val_tracer_pair).unzip2()
    jaxpr, res = jaxpr_trace.to_jaxpr(list(out_tracer_ft), dbg, source_info_util.current())
    in_tree, out_tree = args_ft.tree, out_ft.tree
    del trace, in_tracers, out_tracer_ft
  def f_rem(res, *args):
    args_flat = tree_leaves_checked(in_tree, (args, {}))
    out_flat = core.eval_jaxpr(jaxpr, res, *args_flat)
    return tree_unflatten(out_tree, out_flat)
  return out_ft.unflatten(), Partial(f_rem, map(reduce_precision, res))

class RematTracer(core.Tracer['RematTrace']):
  _trace: RematTrace

  def __init__(self, trace, x, jaxpr_tracer):
    super().__init__(trace, core.typeof(x))
    self.val = x
    self.tracer = jaxpr_tracer

class RematTrace(core.Trace):
  def __init__(self, parent_trace, jaxpr_trace, tag, policy):
    super().__init__()
    self.parent_trace = parent_trace
    self.jaxpr_trace = jaxpr_trace
    self.tag = tag
    self.policy = policy
    self.requires_low = False

  def to_val_tracer_pair(self, x):
    if isinstance(x, RematTracer) and x._trace.tag is self.tag:
      return x.val, x.tracer
    else:
      return x, x

  def process_primitive(self, prim, tracers, params, /):
    in_vals, in_vals2 = unzip2(map(self.to_val_tracer_pair, tracers))
    if prim in rules:
      with core.set_current_trace(self.parent_trace):
        out_primal, rem = rules[prim](self.policy, *in_vals, **params)
      with core.set_current_trace(self.jaxpr_trace):
        out_primal2 = rem(*in_vals2)
    else:
      with core.set_current_trace(self.parent_trace):
        out_primal = prim.bind(*in_vals, **params)
      with core.set_current_trace(self.jaxpr_trace):
        out_primal2 = prim.bind(*in_vals2, **params)
    if prim.multiple_results:
      return map(partial(RematTracer, self), out_primal, out_primal2)
    else:
      return RematTracer(self, out_primal, out_primal2)

def reduce_precision(x):
  if (h := reduce_precision_handlers.get(type(t := core.typeof(x)))):
    return h(t, x)
  return x

rules: dict[core.Primitive, Callable] = {}
reduce_precision_handlers: dict[type, Callable] = {}


def remat_jaxpr(jaxpr, policy):
  return _remat_jaxpr(jaxpr, frozenset(policy))

@weakref_lru_cache
def _remat_jaxpr(jaxpr, policy):
  dbg = jaxpr.jaxpr.debug_info
  fwd_trace = pe.DynamicJaxprTrace(dbg)
  rem_trace = pe.DynamicJaxprTrace(dbg, auto_dce=True)
  tag = core.TraceTag()
  trace = RematTrace(fwd_trace, rem_trace, tag, policy)
  rem_trace.tag = tag
  src = source_info_util.current()

  def new_arg(a):
    return RematTracer(trace, fwd_trace.new_arg(a, src), rem_trace.new_arg(a, src))  # noqa: F821

  tracers = map(new_arg, jaxpr.in_aval_qdds)
  with core.set_current_trace(trace, check_leaks=True):
    ans = core.eval_jaxpr(jaxpr.jaxpr, jaxpr.consts, *tracers)
    out_primals, out_rem = unzip2(map(trace.to_val_tracer_pair, ans))
    del trace, ans, new_arg, tracers

  rem_jaxpr_, rem_consts = rem_trace.to_jaxpr(out_rem, dbg.with_unknown_names(), src)
  rem_jaxpr = pe.close_jaxpr(pe.convert_constvars_jaxpr(rem_jaxpr_))
  rem_trace.invalidate()
  rem_consts = map(partial(fwd_trace.to_jaxpr_tracer, source_info=src), rem_consts)
  fwd_jaxpr_, fwd_consts = fwd_trace.to_jaxpr(
      [*out_primals, *rem_consts], dbg.with_unknown_names(), src)
  fwd_trace.invalidate()
  fwd_jaxpr = core.ClosedJaxpr(fwd_jaxpr_, fwd_consts)
  return fwd_jaxpr, rem_jaxpr, len(rem_consts)



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains registrations for pallas_call on TPU."""

from __future__ import annotations

from collections.abc import Sequence
import dataclasses
import functools
import json
from typing import Any, cast

import jax
from jax import dtypes
from jax._src import core as jax_core
from jax._src import frozen_dict
from jax._src import sharding_impls
from jax._src import tpu_custom_call
from jax._src.interpreters import mlir
from jax._src.lib.mlir import ir
from jax._src.lib.mlir import passmanager
from jax._src.pallas import core as pallas_core
from jax._src.pallas.mosaic import core as tpu_core
from jax._src.pallas.mosaic import lowering
from jax._src.pallas.mosaic import sc_core
from jax._src.pallas.mosaic import sc_lowering
from jax._src.pallas.mosaic import tpu_info
from jax._src.state import types as state_types
from jax.experimental import mosaic
from jax.experimental.mosaic.dialects import tpu


def _maybe_cast_to_int(x: jax.Array | jax_core.AbstractValue):
  """Casts boolean values to integers.

  We perform this cast because Mosaic does not directly support bool values
  for Memrefs. Instead, we load bools as integers and cast them to bools
  after loading from a memref inside of the kernel.
  """
  assert isinstance(
      x, (jax.Array, jax_core.ShapedArray, state_types.AbstractLinVal)
  ) or (
      isinstance(x, jax_core.Tracer)
      and isinstance(x.aval, state_types.AbstractLinVal)
  ), type(x)
  if isinstance(x, jax.Array):
    if dtypes.issubdtype(x.dtype, jax.numpy.bool_):
      return x.astype(lowering.BOOL_MEMREF_TYPE)
    return x
  else:
    if dtypes.issubdtype(x.dtype, jax.numpy.bool_):
      if isinstance(x, state_types.AbstractLinVal):
        raise NotImplementedError  # TODO(mattjj,sharadmv)
      return jax_core.ShapedArray(x.shape, lowering.BOOL_MEMREF_TYPE)
    return x


def _get_memory_space_from_aval(
    out_aval: jax_core.AbstractValue, kernel_type: tpu_core.CoreType
) -> tpu_custom_call.MemorySpace | None:
  if not isinstance(out_aval, jax_core.ShapedArray):
    raise ValueError("Memory spaces not defined for non-ShapedArrays")
  if not isinstance(
      ms := getattr(out_aval, "memory_space", None), tpu_core.MemorySpace
  ):
    return None  # If we are passed a non-TPU memory space, ignore it.
  # If we are passed an aval with an explicit memory space tag, we use it
  # to constrain the memory space.
  match ms:
    case tpu_core.MemorySpace.HBM:
      return tpu_custom_call.MemorySpace.HBM
    case tpu_core.MemorySpace.VMEM:
      return tpu_custom_call.MemorySpace.VMEM
    case tpu_core.MemorySpace.SMEM:
      return tpu_custom_call.MemorySpace.SMEM
    case tpu_core.MemorySpace.SEMAPHORE:
      match kernel_type:
        case tpu_core.CoreType.SC_SCALAR_SUBCORE:
          return tpu_custom_call.MemorySpace.SC_SCALAR_SEMAPHORE_MEM
        case tpu_core.CoreType.TC:
          return tpu_custom_call.MemorySpace.SEMAPHORE_MEM
        case _:
          raise ValueError(f"Invalid kernel type for semaphore: {kernel_type}")
    case tpu_core.MemorySpace.HOST:
      return tpu_custom_call.MemorySpace.HOST
    case _:
      pass
  return None


def _get_memory_spaces_from_avals(
    avals: Sequence[jax_core.AbstractValue], kernel_type: tpu_core.CoreType
) -> tuple[tpu_custom_call.MemorySpace | None, ...] | None:
  memory_spaces = None
  if any(isinstance(aval, jax_core.ShapedArray)
         and not isinstance(aval.memory_space, jax_core.MemorySpace)
         for aval in avals):
    memory_spaces = tuple(
        _get_memory_space_from_aval(aval, kernel_type=kernel_type)
        for aval in avals)
  return memory_spaces


def _resolve_memory_spaces(
    in_avals: Sequence[jax_core.AbstractValue],
    out_avals: Sequence[jax_core.AbstractValue],
    *,
    input_output_aliases: tuple[tuple[int, int], ...],
    kernel_type: tpu_core.CoreType,
) -> tuple[
    tuple[tpu_custom_call.MemorySpace | None, ...] | None,
    tuple[tpu_custom_call.MemorySpace | None, ...] | None,
]:
  output_memory_spaces = _get_memory_spaces_from_avals(
      out_avals, kernel_type=kernel_type
  )
  input_memory_spaces = None
  if any(isinstance(aval, jax_core.ShapedArray)
         and not isinstance(aval.memory_space, jax_core.MemorySpace)
         for aval in in_avals):
    input_memory_spaces = _get_memory_spaces_from_avals(
        in_avals, kernel_type=kernel_type
    )
    # ShapedArrayWithMemorySpace wasn't allowed to escape the Pallas kernel, so
    # it was stripped from the original out_avals. We need to resolve this for
    # outputs aliased to inputs with pltpu.with_memory_space_constraint.
    if input_memory_spaces is not None:
      output_memory_spaces_list: list[tpu_custom_call.MemorySpace | None]
      if output_memory_spaces is None:
        output_memory_spaces_list = [None] * len(out_avals)
      else:
        output_memory_spaces_list = list(output_memory_spaces)
      modified = False
      for in_idx, out_idx in input_output_aliases:
        if (ms := input_memory_spaces[in_idx]) is not None:
          if (out_ms := output_memory_spaces_list[out_idx]) not in (None, ms):
            raise ValueError(
                f"In/out memory space conflict for alias {in_idx}->{out_idx}:"
                f" {ms} != {out_ms}"
            )
          output_memory_spaces_list[out_idx] = ms
          modified = True
      if modified:
        output_memory_spaces = tuple(output_memory_spaces_list)
  if output_memory_spaces is not None:
    input_memory_spaces_list: list[tpu_custom_call.MemorySpace | None]
    if input_memory_spaces is None:
      input_memory_spaces_list = [None] * len(in_avals)
    else:
      input_memory_spaces_list = list(input_memory_spaces)
    modified = False
    for in_idx, out_idx in input_output_aliases:
      if (ms := output_memory_spaces[out_idx]) is not None:
        if (in_ms := input_memory_spaces_list[in_idx]) not in (None, ms):
          raise ValueError(
              f"In/out memory space conflict for alias {in_idx}->{out_idx}:"
              f"{in_ms} != {ms}"
          )
        input_memory_spaces_list[in_idx] = ms
        modified = True
    if modified:
      input_memory_spaces = tuple(input_memory_spaces_list)
  if input_memory_spaces is not None:
    input_memory_spaces = tuple(
        i
        if i
        in {
            tpu_custom_call.MemorySpace.HBM,
            tpu_custom_call.MemorySpace.VMEM,
            tpu_custom_call.MemorySpace.SMEM,
        }
        else None
        for i in input_memory_spaces
    )
  return input_memory_spaces, output_memory_spaces


def _resolve_side_effect_type(
    has_side_effects: bool | tpu_core.SideEffectType,
) -> bool | tpu_custom_call.TpuSideEffectType:
  match has_side_effects:
    case bool():
      return has_side_effects
    case tpu_core.SideEffectType.PURE:
      return tpu_custom_call.TpuSideEffectType.PURE
    case tpu_core.SideEffectType.DATAFLOW_SIDE_EFFECTING:
      return tpu_custom_call.TpuSideEffectType.DATAFLOW_SIDE_EFFECTING
    case tpu_core.SideEffectType.SIDE_EFFECTING:
      return tpu_custom_call.TpuSideEffectType.SIDE_EFFECTING
    case _:
      raise ValueError(f"Invalid side effect type: {has_side_effects}")


def _resolve_tiling(
    mosaic_params: tpu_core.CompilerParams,
) -> tpu_custom_call.Tiling | None:
  if mosaic_params.use_tc_tiling_on_sc is None:
    return None
  if mosaic_params.kernel_type is tpu_core.CoreType.TC:
    raise ValueError(
        "use_tc_tiling_on_sc= is not supported for TC kernels"
    )

  return (
      tpu_custom_call.Tiling.COMPACT
      if mosaic_params.use_tc_tiling_on_sc
      else tpu_custom_call.Tiling.SPARSE_CORE
  )


def _lower_to_custom_call(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    mosaic_module: ir.Module,
    mosaic_params: tpu_core.CompilerParams,
    num_dynamic_grid_bounds: int,
    input_output_aliases: tuple[tuple[int, int], ...],
    cost_estimate: pallas_core.CostEstimate | None,
    out_avals: tuple[jax_core.AbstractValue, ...],
    effects: jax_core.Effects,
    metadata: frozen_dict.FrozenDict[str, str] | None,
    name: str,
    jax_mesh,
):
  kernel_type = mosaic_params.kernel_type
  input_output_aliases = tuple(
      (a[0] + num_dynamic_grid_bounds, a[1]) for a in input_output_aliases
  )

  # Replace in_avals to physical avals.
  # This step is required for mapping logical types to physical types.
  # (e.g. PRNG key -> uint32[2])
  physical_avals = [jax_core.physical_aval(aval) for aval in ctx.avals_in]
  ctx = ctx.replace(avals_in=physical_avals)

  # Booleans are loaded into the kernel as integers.
  def _maybe_cast_inputs(*args):
    args = [_maybe_cast_to_int(x) for x in args]
    return args

  kernel_in_avals = [_maybe_cast_to_int(x) for x in ctx.avals_in]
  kernel_out_avals = [_maybe_cast_to_int(x) for x in ctx.avals_out]
  cast_ctx = ctx.replace(avals_out=kernel_in_avals)
  in_nodes = mlir.lower_fun(_maybe_cast_inputs)(cast_ctx, *in_nodes)

  # Dynamic grid bounds have to go at the front.
  dynamic_grid_args, args = (
      in_nodes[:num_dynamic_grid_bounds],
      in_nodes[num_dynamic_grid_bounds:],
  )
  kernel_ctx = ctx.replace(avals_in=kernel_in_avals, avals_out=kernel_out_avals)
  input_memory_spaces, output_memory_spaces = _resolve_memory_spaces(
      ctx.avals_in,
      out_avals,
      input_output_aliases=input_output_aliases,
      kernel_type=kernel_type,
  )

  if cost_estimate is not None:
    mosaic_cost_estimate = cast(
        tpu_custom_call.CostEstimate, dataclasses.asdict(cost_estimate)
    )
  else:
    mosaic_cost_estimate = None

  dict_metadata = dict(metadata) if metadata is not None else {}
  del metadata
  if jax_mesh is not None:
    mesh_axes = {
        e.name
        for e in effects
        if isinstance(e, jax_core.NamedAxisEffect)
        # Filter for only device mesh axis name effects
        and e.name in jax_mesh.axis_names
    }
    # Only put mesh axes in metadata if there are any.
    if mesh_axes:
      if "mesh_axes" in dict_metadata:
        raise ValueError("Metadata already contains mesh axes.")
      mesh_axes_list = list(mesh_axes)
      if all(isinstance(a, str) for a in mesh_axes):
        mesh_axes_list = sorted(mesh_axes)  # pyrefly: ignore[bad-specialization]
      dict_metadata["mesh_axes"] = json.dumps(mesh_axes_list)
  out_nodes = mosaic.lower_module_to_custom_call(
      kernel_ctx,
      *dynamic_grid_args,
      *args,
      module=mosaic_module,
      out_type=kernel_out_avals,
      kernel_name=mlir.sanitize_name(name),
      cost_estimate=mosaic_cost_estimate,
      vmem_limit_bytes=mosaic_params.vmem_limit_bytes,
      flags=mosaic_params.flags,
      allow_input_fusion=mosaic_params.allow_input_fusion,
      input_output_aliases=input_output_aliases,
      serialization_format=mosaic_params.serialization_format,
      internal_scratch_in_bytes=mosaic_params.internal_scratch_in_bytes,
      collective_id=mosaic_params.collective_id,
      has_side_effects=_resolve_side_effect_type(
          mosaic_params.has_side_effects
      ),
      output_memory_spaces=output_memory_spaces,
      disable_bounds_checks=mosaic_params.disable_bounds_checks,
      disable_semaphore_checks=mosaic_params.disable_semaphore_checks,
      input_memory_spaces=input_memory_spaces,
      metadata=dict_metadata,
      skip_device_barrier=mosaic_params.skip_device_barrier,
      allow_collective_id_without_custom_barrier=mosaic_params.allow_collective_id_without_custom_barrier,
      shape_invariant_numerics=mosaic_params.shape_invariant_numerics,
      tiling=_resolve_tiling(mosaic_params),
  )
  _maybe_cast_to_bool = (
      lambda x, aval: x.astype(jax.numpy.bool_)
      if aval.dtype == jax.numpy.bool_
      else x
  )

  def _maybe_cast_outputs(*args):
    args = [_maybe_cast_to_bool(x, aval) for x, aval in zip(args, out_avals)]
    return args

  cast_ctx = ctx.replace(avals_in=kernel_out_avals)
  return mlir.lower_fun(_maybe_cast_outputs)(cast_ctx, *out_nodes)


def pallas_call_tpu_lowering_rule(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    jaxpr: jax_core.Jaxpr,
    grid_mapping: pallas_core.GridMapping,
    mesh: pallas_core.Mesh | None,
    input_output_aliases: tuple[tuple[int, int], ...],
    debug: bool,
    interpret: bool,
    compiler_params: pallas_core.CompilerParams | None,
    cost_estimate: pallas_core.CostEstimate | None,
    out_avals: tuple[jax_core.AbstractValue, ...],
    metadata: frozen_dict.FrozenDict[str, str] | None,
    name: str | None,
):
  """Lowers a pallas_call to a Mosaic TPU custom call."""
  del interpret  # Unused.

  debug_info = jaxpr.debug_info
  if debug:
    print(f"\nThe kernel jaxpr for pallas_call {debug_info.func_src_info}:")
    print(jaxpr)

  if compiler_params is None:
    mosaic_params = tpu_core.CompilerParams()
  else:
    assert isinstance(compiler_params, tpu_core.CompilerParams)
    mosaic_params = compiler_params

  kernel_type = mosaic_params.kernel_type

  # `mesh` argument is the core mesh if provided by the user (e.g. in core_map).
  # If it's None, we create a default mesh based on the kernel type.
  # TODO(rdyro): Remove once we have a way of explicitly passing a mesh here.
  if mesh is None:
    if kernel_type == tpu_core.CoreType.TC:
      # TODO(rdyro): In cross-compilation, TPU info might not be available.
      # Remove this once we always have an explicit mesh.
      try:
        num_cores = tpu_info.get_tpu_info().num_cores
      except ValueError:
        num_cores = 1
      mesh = tpu_core.create_tensorcore_mesh(
          axis_name="tensorcore_unnamed_core", num_cores=num_cores
      )
    elif kernel_type == tpu_core.CoreType.SC_SCALAR_SUBCORE:
      mesh = sc_core.ScalarSubcoreMesh(axis_name="sparsecore_unnamed_core")
    elif kernel_type == tpu_core.CoreType.SC_VECTOR_SUBCORE:
      mesh = sc_core.VectorSubcoreMesh(
          core_axis_name="sparsecore_unnamed_core",
          subcore_axis_name="sparsecore_unnamed_subcore",
      )
    else:
      raise ValueError(f"Unsupported kernel type: {kernel_type}")
  mpmd_meshes = {kernel_type: mesh}

  jax_mesh = None
  axis_context = ctx.module_context.axis_context
  if axis_context is not None:
    if isinstance(axis_context, sharding_impls.SPMDAxisContext):
      jax_mesh = axis_context.mesh
  mlir_ctx = mlir.JaxIrContext()
  mlir_ctx.append_dialect_registry(mlir.upstream_dialects)
  mlir_ctx.load_all_available_dialects()
  tpu.register_dialect(mlir_ctx)

  match kernel_type:
    case tpu_core.CoreType.TC:
      lower_jaxpr_to_module = lowering.lower_jaxpr_to_module
    case (
        tpu_core.CoreType.SC_SCALAR_SUBCORE
        | tpu_core.CoreType.SC_VECTOR_SUBCORE
    ):
      lower_jaxpr_to_module = functools.partial(
          sc_lowering.lower_pipelined_jaxpr_to_module,
          use_tc_tiling=mosaic_params.use_tc_tiling_on_sc,
      )
    case _:
      raise ValueError(f"Unsupported kernel type: {mosaic_params.kernel_type}")

  with mlir_ctx, ir.Location.unknown(mlir_ctx):
    mosaic_module = lower_jaxpr_to_module(
        ctx,
        grid_mapping,
        jaxpr,
        dimension_semantics=mosaic_params.dimension_semantics,
        kernel_type=kernel_type,
        mesh=jax_mesh,
        dynamic_shape_replacement_enabled=pallas_core.dynamic_shapes_export_enabled(),
        mpmd_meshes=mpmd_meshes,
    )

  if debug:
    pm = passmanager.PassManager.parse("builtin.module(canonicalize)", mlir_ctx)
    pm.run(mosaic_module.operation)
    print(f"\nThe Mosaic module for pallas_call {debug_info.func_src_info}:")
    print(mosaic_module)

  return _lower_to_custom_call(
      ctx,
      *in_nodes,
      mosaic_module=mosaic_module,
      mosaic_params=mosaic_params,
      num_dynamic_grid_bounds=grid_mapping.num_dynamic_grid_bounds,
      input_output_aliases=input_output_aliases,
      cost_estimate=cost_estimate,
      out_avals=out_avals,
      effects=jaxpr.effects,
      metadata=metadata,
      name=name or debug_info.func_name,
      jax_mesh=jax_mesh,
  )


def mpmd_map_tpu_lowering_rule(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    meshes,
    jaxprs,
    grid_mappings,
    out_avals,
    input_output_aliases,
    compiler_params,
    interpret,
    debug,
    cost_estimate,
    metadata,
    name,
):
  del interpret  # Unused.

  if debug:
    for idx, jaxpr in enumerate(jaxprs):
      print(
          f"\nThe kernel jaxpr {idx=} for mpmd_map"
          f" {jaxpr.debug_info.func_src_info}:"
      )
      print(jaxpr)

  if compiler_params is None:
    mosaic_params = tpu_core.CompilerParams()
  else:
    assert isinstance(compiler_params, tpu_core.CompilerParams)
    mosaic_params = compiler_params

  # TODO(slebedev): Check kernel type and raise if it is set.
  if mosaic_params.dimension_semantics is not None:
    raise ValueError(
        "mpmd_map does not support dimension_semantics= in compiler_params="
    )

  mpmd_meshes_map = {mesh.kernel_type: mesh for mesh in meshes}
  jax_mesh = None
  axis_context = ctx.module_context.axis_context
  if axis_context is not None:
    if isinstance(axis_context, sharding_impls.SPMDAxisContext):
      jax_mesh = axis_context.mesh
  mlir_ctx = mlir.JaxIrContext()
  mlir_ctx.append_dialect_registry(mlir.upstream_dialects)
  mlir_ctx.load_all_available_dialects()
  tpu.register_dialect(mlir_ctx)

  with mlir_ctx, ir.Location.unknown(mlir_ctx):
    mosaic_module = ir.Module.create()
    for mesh, jaxpr, grid_mapping in zip(
        meshes, jaxprs, grid_mappings, strict=True
    ):
      if (
          not hasattr(mesh, "kernel_type") or
          not hasattr(mesh, "dimension_semantics")
      ):
        raise ValueError(
            "mpmd_map requires the mesh to define its ``kernel_type`` and"
            " ``dimension_semantics``"
        )

      match kernel_type := mesh.kernel_type:
        case tpu_core.CoreType.TC:
          if (mpmd_meshes_map is not None
              and mpmd_meshes_map.keys() != {tpu_core.CoreType.TC}):
            raise NotImplementedError(
                "mpmd_map does not support TC kernels yet.")
          lower_fn = lowering.lower_jaxpr_into_module
        case (
            tpu_core.CoreType.SC_SCALAR_SUBCORE
            | tpu_core.CoreType.SC_VECTOR_SUBCORE
        ):
          lower_fn = sc_lowering.lower_jaxpr_into_module
        case _:
          raise ValueError(
              f"Unsupported kernel type: {kernel_type}"
          )

      lower_fn(
          ctx,
          mosaic_module,
          grid_mapping,
          jaxpr,
          dimension_semantics=mesh.dimension_semantics,
          kernel_type=kernel_type,
          mesh=jax_mesh,
          name=mlir.sanitize_name(jaxpr.debug_info.func_name),
          dynamic_shape_replacement_enabled=pallas_core.dynamic_shapes_export_enabled(),
          mpmd_meshes=mpmd_meshes_map,
      )

  if debug:
    pm = passmanager.PassManager.parse("builtin.module(canonicalize)", mlir_ctx)
    pm.run(mosaic_module.operation)
    print("\nThe Mosaic module for mpmd_map:")
    print(mosaic_module)

  if name is None:
    name = "_".join(jaxpr.debug_info.func_name for jaxpr in jaxprs)

  match [*{mesh.kernel_type for mesh in meshes}]:
    case [kernel_type]:
      mosaic_params = dataclasses.replace(
          mosaic_params, kernel_type=kernel_type
      )
    case _:
      # Use a stub ``kernel_type`` if we are lowering multiple kernels.
      # This will hopefully cause a runtime error if ``kernel_type`` is ever
      # accessed.
      mosaic_params = dataclasses.replace(
          mosaic_params, kernel_type=cast(Any, object())
      )

  return _lower_to_custom_call(
      ctx,
      *in_nodes,
      mosaic_module=mosaic_module,
      mosaic_params=mosaic_params,
      num_dynamic_grid_bounds=0,
      input_output_aliases=tuple(input_output_aliases.items()),
      cost_estimate=cost_estimate,
      out_avals=out_avals,
      effects=set().union(*(jaxpr.effects for jaxpr in jaxprs)),
      metadata=metadata,
      name=name,
      jax_mesh=jax_mesh,
  )


pallas_core.register_lowering_rule(tpu_core.CompilerParams, pallas_call_tpu_lowering_rule, "tpu")



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Serialization and deserialization of _export.Exported

from __future__ import annotations


from collections.abc import Callable, Iterable
import dataclasses
import itertools
from functools import partial
import types
from typing import cast, Any, TypeVar

try:
  import flatbuffers
except ImportError as e:
  raise ImportError(
      "Please install 'flatbuffers' in order to use Exported serialization"
      ) from e

from jax._src import core
from jax._src import dtypes
from jax._src import effects
from jax._src.export import serialization_generated as ser_flatbuf
from jax._src.export import _export
from jax._src.export import shape_poly
from jax._src.lib import xla_client
from jax._src import mesh
from jax._src import named_sharding
from jax._src import partition_spec
from jax._src import tree_util

import numpy as np

T = TypeVar("T")
SerT = TypeVar("SerT")

# The _SERIALIZATION_VERSION changes when we change the serialization schema
# even if the change is backwards compatible.
# Version 1, Nov 2023, first version.
# Version 2, Dec 16th, 2023, adds the f0 dtype.
# Version 3, October 16th, 2024, adds serialization for namedtuple and custom types
#   This version is backwards compatible with Version 2.
# Version 4, April 7th, 2025, adds serialization for PRNGs key types.
#   This version is backwards compatible with Version 2 and 3.
# Version 5, November 23rd, 2025, adds serialization for aval memory_space,
#   upgrade num_devices to a 32 bit value.
#   This version is backwards compatible with Version 2 to 4.
# Version 6, January 15th, 2026, adds serialization for sharding as
#   NamedSharding, including the abstract mesh, and the partition spec.
#   Contains also HloSharding serialization, for forward compatibility.
#   This version is backwards compatible with Version 2 to 5.
# Version 7 was briefly live but pulled back due to breaking compatiblity.
# Version 8, March 12th, 2026, add serializaton for AbstractMesh.abstract_device.
#   This version is backwards compatible with Version 2 to 7.
# Version 9, March 17th, 2026, removes HloSharding serialization.
#   This is another attempt at what Version 7 was supposed to be.
#   This version is backwards compatible with Version 2 to 8.
# Version 10, April 4th, 2026, optimizes serialization of duplicate shardings,
#   abstract meshes and avals.
_SERIALIZATION_VERSION = 10


@dataclasses.dataclass
class _SerializedUniques:
  # Map unique objects to their index in the serialized data.
  unique_avals: list[core.AbstractValue]
  avals_map: dict[core.AbstractValue, int]
  unique_abstract_meshes: list[mesh.AbstractMesh]
  abstract_meshes_map: dict[mesh.AbstractMesh, int]
  unique_named_shardings: list[named_sharding.NamedSharding]
  named_shardings_map: dict[named_sharding.NamedSharding, int]

  @staticmethod
  def create_from_exported(exp: _export.Exported):
    uniques = _SerializedUniques([], {}, [], {}, [], {})
    for aval in itertools.chain(exp.in_avals, exp.out_avals):
      uniques.add_aval(aval)
    for sharding in itertools.chain(exp._in_named_shardings,
                                    exp._out_named_shardings):
      uniques.add_named_sharding(sharding)
    return uniques

  @staticmethod
  def create_from_uniques(unique_avals: list[core.AbstractValue],
                          unique_abstract_meshes: list[mesh.AbstractMesh],
                          unique_named_shardings: list[named_sharding.NamedSharding]):
    uniques = _SerializedUniques([], {}, [], {}, [], {})
    uniques.unique_avals = unique_avals
    uniques.avals_map = {a: i for i, a in enumerate(unique_avals)}
    uniques.unique_abstract_meshes = unique_abstract_meshes
    uniques.abstract_meshes_map = {m: i for i, m in enumerate(unique_abstract_meshes)}
    uniques.unique_named_shardings = unique_named_shardings
    uniques.named_shardings_map = {s: i for i, s in enumerate(unique_named_shardings)}
    return uniques

  def add_aval(self, aval: core.AbstractValue):
    if aval not in self.avals_map:
      self.avals_map[aval] = len(self.unique_avals)
      self.unique_avals.append(aval)

  def add_named_sharding(self, sharding: named_sharding.NamedSharding | None):
    if sharding is None:
      return
    amesh = sharding.mesh.abstract_mesh
    if amesh is not None and amesh not in self.abstract_meshes_map:
      self.abstract_meshes_map[amesh] = len(self.unique_abstract_meshes)
      self.unique_abstract_meshes.append(amesh)
    if sharding not in self.named_shardings_map:
      self.named_shardings_map[sharding] = len(self.unique_named_shardings)
      self.unique_named_shardings.append(sharding)


def serialize(exp: _export.Exported, vjp_order: int = 0) -> bytearray:
  """Serializes an Exported.

  Args:
    exp: the Exported to serialize.
    vjp_order: The maximum vjp order to include. E.g., the value 2 means that we
      serialize the primal functions and two orders of the `vjp` function. This
      should allow 2nd order reverse mode differentiation of the deserialized
      function. i.e., `jax.grad(jax.grad(f)).`
  """
  builder = flatbuffers.Builder(65536)
  exported = _serialize_exported(builder, exp, vjp_order)
  builder.Finish(exported)
  return builder.Output()


def deserialize(ser: bytearray) -> _export.Exported:
  """Deserializes an Exported."""
  exp = ser_flatbuf.Exported.GetRootAsExported(ser)
  return _deserialize_exported(exp)


def _serialize_exported(
    builder: flatbuffers.Builder, exp: _export.Exported, vjp_order: int
) -> int:
  uniques = _SerializedUniques.create_from_exported(exp)
  if not exp._has_named_shardings:
    raise ValueError(
      "Exported being serialized must have named shardings after 3/17/2026.")
  # Serialize bottom-up
  fun_name = builder.CreateString(exp.fun_name)
  in_tree = _serialize_pytreedef(builder, exp.in_tree)
  # TODO(necula): stop serializing in_avals 1 month after 4/4/26.
  in_avals = _serialize_array(builder, _serialize_aval, exp.in_avals)

  out_tree = _serialize_pytreedef(builder, exp.out_tree)
  # TODO(necula): stop serializing out_avals 1 month after 4/4/26
  out_avals = _serialize_array(builder, _serialize_aval, exp.out_avals)
  # TODO(necula): stop serializing in_shardings 1 month after 4/4/26
  in_shardings = _serialize_array(
      builder, partial(_serialize_sharding, uniques=uniques),
      exp._in_named_shardings)
  # TODO(necula): stop serializing out_shardings 1 month after 4/4/26
  out_shardings = _serialize_array(
      builder, partial(_serialize_sharding, uniques=uniques),
      exp._out_named_shardings)
  ordered_effects = _serialize_array(
      builder, _serialize_effect, exp.ordered_effects
  )
  unordered_effects = _serialize_array(
      builder, _serialize_effect, exp.unordered_effects
  )
  disabled_safety_checks = _serialize_array(
      builder, _serialize_disabled_safety_check, exp.disabled_safety_checks
  )
  platforms = _serialize_array(
      builder, lambda b, p: b.CreateString(p), exp.platforms
  )
  mlir_module_serialized = builder.CreateByteVector(exp.mlir_module_serialized)
  module_kept_var_idx = builder.CreateNumpyVector(
      np.array(exp.module_kept_var_idx, dtype=np.uint16)
  )

  vjp = None
  if vjp_order > 0:
    if not exp.has_vjp():
      # TODO: add test
      raise ValueError(
          "serialization of an Exported that does not have vjps of high-enough "
          "order"
      )
    vjp = _serialize_exported(builder, exp.vjp(), vjp_order - 1)

  unique_avals_offset = _serialize_array(
      builder, _serialize_aval, uniques.unique_avals)
  unique_abstract_meshes_offset = _serialize_array(
      builder, _serialize_abstract_mesh, uniques.unique_abstract_meshes)
  unique_named_shardings_offset = _serialize_array(
      builder, partial(_serialize_named_sharding, uniques=uniques),
      uniques.unique_named_shardings)

  in_aval_idxs = builder.CreateNumpyVector(
    np.array([uniques.avals_map[a] for a in exp.in_avals], dtype=np.uint32))
  out_aval_idxs = builder.CreateNumpyVector(
    np.array([uniques.avals_map[a] for a in exp.out_avals], dtype=np.uint32))

  in_shardings_idxs = builder.CreateNumpyVector(
    np.array([0 if s is None else 1 + uniques.named_shardings_map[s]
              for s in exp._in_named_shardings], dtype=np.uint32))
  out_shardings_idxs = builder.CreateNumpyVector(
    np.array([0 if s is None else 1 + uniques.named_shardings_map[s]
              for s in exp._out_named_shardings], dtype=np.uint32))

  ser_flatbuf.ExportedStart(builder)
  # TODO(necula): we cannot really store the actual serialization_version
  # in the flatbuffer because prior to 11/25/2025 deserializers checked
  # if the version is 2 or 3. I have now removed that check, but for the
  # sake of old deserializers we can only store version 3. Starting
  # on January 2026 we can store the actual version.
  ser_flatbuf.ExportedAddSerializationVersion(builder, 3)
  ser_flatbuf.ExportedAddFunctionName(builder, fun_name)
  ser_flatbuf.ExportedAddInTree(builder, in_tree)
  ser_flatbuf.ExportedAddInAvals(builder, in_avals)
  ser_flatbuf.ExportedAddOutTree(builder, out_tree)
  ser_flatbuf.ExportedAddOutAvals(builder, out_avals)
  ser_flatbuf.ExportedAddNrDevices(builder, exp.nr_devices)
  ser_flatbuf.ExportedAddInShardings(builder, in_shardings)
  ser_flatbuf.ExportedAddOutShardings(builder, out_shardings)
  ser_flatbuf.ExportedAddPlatforms(builder, platforms)
  ser_flatbuf.ExportedAddOrderedEffects(builder, ordered_effects)
  ser_flatbuf.ExportedAddUnorderedEffects(builder, unordered_effects)
  ser_flatbuf.ExportedAddDisabledChecks(builder, disabled_safety_checks)
  ser_flatbuf.ExportedAddMlirModuleSerialized(builder, mlir_module_serialized)
  ser_flatbuf.ExportedAddCallingConventionVersion(
      builder, exp.calling_convention_version
  )
  ser_flatbuf.ExportedAddModuleKeptVarIdx(builder, module_kept_var_idx)
  ser_flatbuf.ExportedAddUsesGlobalConstants(
      builder, exp.uses_global_constants
  )
  if vjp is not None:
    ser_flatbuf.ExportedAddVjp(builder, vjp)

  ser_flatbuf.ExportedAddUniqueAvals(builder, unique_avals_offset)
  ser_flatbuf.ExportedAddUniqueAbstractMeshes(builder,
                                              unique_abstract_meshes_offset)
  ser_flatbuf.ExportedAddUniqueNamedShardings(builder,
                                              unique_named_shardings_offset)
  ser_flatbuf.ExportedAddInAvalsIdxs(builder, in_aval_idxs)
  ser_flatbuf.ExportedAddOutAvalsIdxs(builder, out_aval_idxs)
  ser_flatbuf.ExportedAddInShardingsIdxs(builder, in_shardings_idxs)
  ser_flatbuf.ExportedAddOutShardingsIdxs(builder, out_shardings_idxs)

  return ser_flatbuf.ExportedEnd(builder)


def _serialize_array(
    builder: flatbuffers.Builder,
    serialize_one: Callable[[flatbuffers.Builder, T], int],
    elements: Iterable[T],
) -> int:
  element_offsets = [serialize_one(builder, e) for e in elements]
  del elements
  ser_flatbuf.PyTreeDefStartChildrenVector(builder, len(element_offsets))
  for sc in reversed(element_offsets):
    builder.PrependUOffsetTRelative(sc)
  return builder.EndVector()


def _deserialize_exported(exp: ser_flatbuf.Exported) -> _export.Exported:
  scope = shape_poly.SymbolicScope(())  # TODO(necula): serialize the constraints

  unique_avals = [
      _deserialize_aval(exp.UniqueAvals(i), scope=scope, sharding=None)
      for i in range(exp.UniqueAvalsLength())]
  unique_abstract_meshes = [
      _deserialize_abstract_mesh(exp.UniqueAbstractMeshes(i))
      for i in range(exp.UniqueAbstractMeshesLength())
  ]
  uniques = _SerializedUniques.create_from_uniques(unique_avals,  # pyrefly: ignore[bad-argument-type]
                                                   unique_abstract_meshes,
                                                   [])
  unique_named_shardings = [
      _deserialize_named_sharding(exp.UniqueNamedShardings(i), uniques=uniques)
      for i in range(exp.UniqueNamedShardingsLength())
  ]
  uniques = _SerializedUniques.create_from_uniques(unique_avals,  # pyrefly: ignore[bad-argument-type]
                                                   unique_abstract_meshes,
                                                   unique_named_shardings)

  fun_name = exp.FunctionName().decode("utf-8")
  _, in_tree = tree_util.tracing_registry.flatten(
      _deserialize_pytreedef_to_pytree(exp.InTree())
  )
  _, out_tree = tree_util.tracing_registry.flatten(
      _deserialize_pytreedef_to_pytree(exp.OutTree())
  )

  # TODO(necula): remove the fallback to NrDevicesShort and mark
  # the field "deprecated" once we abandon the old
  # serialization format (6 months after 11/24/2025).
  nr_devices = exp.NrDevices() or exp.NrDevicesShort()
  def sharding_by_idx(idx):
    if idx == 0:
      return None
    return uniques.unique_named_shardings[idx - 1]

  if exp.InShardingsIdxsLength() > 0:
    in_shardings = tuple(
        sharding_by_idx(exp.InShardingsIdxs(i))
        for i in range(exp.InShardingsIdxsLength())
    )
  elif exp.InShardingsLength() > 0:
    # TODO(necula): remove 6 months after 4/4/26
    in_shardings = tuple(
        _deserialize_sharding(exp.InShardings(i), uniques=uniques)
        for i in range(exp.InShardingsLength())
    )
  else:
    in_shardings = ()

  if exp.OutShardingsIdxsLength() > 0:
    out_shardings = tuple(
      sharding_by_idx(exp.OutShardingsIdxs(i))
      for i in range(exp.OutShardingsIdxsLength())
    )
  elif exp.OutShardingsLength() > 0:
    # TODO(necula): remove 6 months after 4/4/26
    out_shardings = tuple(
      _deserialize_sharding(exp.OutShardings(i), uniques=uniques)
      for i in range(exp.OutShardingsLength())
    )
  else:
    out_shardings = ()

  # has_named_sharding will be True for all exports created after 1/15/2026
  # TODO(b/489569164): remove has_named_sharding 6 months after 1/15/2026
  has_named_shardings = not any(isinstance(s, _export.HloSharding)
                                for s in itertools.chain(in_shardings, out_shardings))
  if has_named_shardings:
    def get_aval_by_idx(idx, sharding: _export.NamedSharding | None):
      base_aval = uniques.unique_avals[idx]
      if sharding is None:
        return base_aval
      return core.update_aval_with_sharding(base_aval, sharding)

    if exp.InAvalsIdxsLength() > 0:
      in_avals = tuple(
          get_aval_by_idx(exp.InAvalsIdxs(i), in_shardings[i])  # pyrefly: ignore[bad-argument-type]
          for i in range(exp.InAvalsIdxsLength()))
    elif exp.InAvalsLength() > 0:
      # TODO(necula): remove 6 months after 4/4/26
      in_avals = tuple(
          _deserialize_aval(exp.InAvals(i), scope=scope, sharding=in_shardings[i])  # pyrefly: ignore[bad-argument-type]
          for i in range(exp.InAvalsLength()))
    else:
      in_avals = ()

    if exp.OutAvalsIdxsLength() > 0:
      out_avals = tuple(
          get_aval_by_idx(exp.OutAvalsIdxs(i), out_shardings[i])  # pyrefly: ignore[bad-argument-type]
                          for i in range(exp.OutAvalsIdxsLength()))
    elif exp.OutAvalsLength() > 0:
      # TODO(necula): remove 6 months after 4/4/26
      out_avals = tuple(
        _deserialize_aval(exp.OutAvals(i), scope=scope, sharding=out_shardings[i])  # pyrefly: ignore[bad-argument-type]
        for i in range(exp.OutAvalsLength())
      )
    else:
      out_avals = ()

    in_shardings_hlo = tuple(_export.named_to_hlo_sharding(s, aval)  # pyrefly: ignore[bad-argument-type]
                             for s, aval in zip(in_shardings, in_avals))
    out_shardings_hlo = tuple(_export.named_to_hlo_sharding(s, aval)  # pyrefly: ignore[bad-argument-type]
                             for s, aval in zip(out_shardings, out_avals))
  else:
    # Export from before 1/15/26
    in_avals = tuple(
        _deserialize_aval(exp.InAvals(i), scope=scope, sharding=None)
        for i in range(exp.InAvalsLength())
    )
    out_avals = tuple(
        _deserialize_aval(exp.OutAvals(i), scope=scope, sharding=None)
        for i in range(exp.OutAvalsLength())
    )
    in_shardings_hlo = cast(tuple[_export.HloSharding | None, ...], in_shardings)
    in_shardings = (None,) * len(in_shardings)
    out_shardings_hlo = cast(tuple[_export.HloSharding | None, ...], out_shardings)
    out_shardings = (None,) * len(out_shardings)

  platforms = tuple(
      exp.Platforms(i).decode("utf-8")
      for i in range(exp.PlatformsLength())
  )
  ordered_effects = tuple(
      _deserialize_effect(exp.OrderedEffects(i))
      for i in range(exp.OrderedEffectsLength())
  )
  unordered_effects = tuple(
      _deserialize_effect(exp.UnorderedEffects(i))
      for i in range(exp.UnorderedEffectsLength())
  )
  disabled_safety_checks = tuple(
      _deserialize_disabled_safety_check(exp.DisabledChecks(i))
      for i in range(exp.DisabledChecksLength())
  )

  mlir_module_serialized = exp.MlirModuleSerializedAsNumpy().tobytes()
  calling_convention_version = exp.CallingConventionVersion()
  module_kept_var_idx = tuple(exp.ModuleKeptVarIdxAsNumpy().tolist())
  uses_global_constants = exp.UsesGlobalConstants()

  _get_vjp = None
  if vjp := exp.Vjp():
    _get_vjp = lambda _: _deserialize_exported(vjp)

  return _export.Exported(
      fun_name=fun_name,
      in_tree=in_tree,
      in_avals=in_avals,
      out_tree=out_tree,
      out_avals=out_avals,
      nr_devices=nr_devices,
      in_shardings_hlo=in_shardings_hlo,
      out_shardings_hlo=out_shardings_hlo,
      _has_named_shardings=has_named_shardings,
      _in_named_shardings=in_shardings,  # pyrefly: ignore[bad-argument-type]
      _out_named_shardings=out_shardings,  # pyrefly: ignore[bad-argument-type]
      platforms=platforms,
      ordered_effects=ordered_effects,
      unordered_effects=unordered_effects,
      disabled_safety_checks=disabled_safety_checks,
      mlir_module_serialized=mlir_module_serialized,
      calling_convention_version=calling_convention_version,
      module_kept_var_idx=module_kept_var_idx,
      uses_global_constants=uses_global_constants,
      _get_vjp=_get_vjp,
  )



def _serialize_pytreedef(
    builder: flatbuffers.Builder, p: tree_util.PyTreeDef
) -> int:
  node_data = p.node_data()
  children = p.children()

  children_vector_offset = None
  children_names_vector_offset = None
  if children:
    children_vector_offset = _serialize_array(
        builder, _serialize_pytreedef, children
    )
  custom_name = None
  custom_auxdata = None
  node_type = node_data and node_data[0]

  if node_data is None:  # leaf
    kind = ser_flatbuf.PyTreeDefKind.leaf
  elif node_type is types.NoneType:
    kind = ser_flatbuf.PyTreeDefKind.none
  elif node_type is tuple:
    kind = ser_flatbuf.PyTreeDefKind.tuple
  elif node_type is list:
    kind = ser_flatbuf.PyTreeDefKind.list
  elif node_type is dict:
    kind = ser_flatbuf.PyTreeDefKind.dict
    assert len(node_data[1]) == len(children)
    def serialize_key(builder, k):
      if not isinstance(k, str):
        raise TypeError(
            "Serialization is supported only for dictionaries with string keys."
            f" Found key {k} of type {type(k)}.")
      return builder.CreateString(k)
    children_names_vector_offset = _serialize_array(
        builder, serialize_key, node_data[1]
    )
  elif node_type in _export.serialization_registry:
    assert node_type is not None
    kind = ser_flatbuf.PyTreeDefKind.custom
    serialized_name, serialize_auxdata = _export.serialization_registry[node_type]
    custom_name = builder.CreateString(serialized_name)
    serialized_auxdata = serialize_auxdata(node_data[1])
    if not isinstance(serialized_auxdata, (bytes, bytearray)):
      raise ValueError(
          "The custom serialization function for `node_type` must "
          f"return a `bytes` object. It returned a {type(serialized_auxdata)}.")
    custom_auxdata = builder.CreateByteVector(serialized_auxdata)
  else:
    raise ValueError(
        "Cannot serialize PyTreeDef containing an "
        f"unregistered type `{node_type}`. "
        "Use `export.register_pytree_node_serialization` or "
        "`export.register_namedtuple_serialization`.")

  ser_flatbuf.PyTreeDefStart(builder)
  ser_flatbuf.PyTreeDefAddKind(builder, kind)
  if children_vector_offset:
    ser_flatbuf.PyTreeDefAddChildren(builder, children_vector_offset)
  if children_names_vector_offset:
    ser_flatbuf.PyTreeDefAddChildrenNames(builder, children_names_vector_offset)
  if custom_name is not None:
    ser_flatbuf.PyTreeDefAddCustomName(builder, custom_name)
  if custom_auxdata is not None:
    ser_flatbuf.PyTreeDefAddCustomAuxdata(builder, custom_auxdata)
  return ser_flatbuf.PyTreeDefEnd(builder)


def _deserialize_pytreedef_to_pytree(p: ser_flatbuf.PyTreeDef):
  # We construct a PyTree and later we'll flatten it to get the PyTreeDef.
  # TODO: is there a more direct way to construct a PyTreeDef?
  kind = p.Kind()
  nr_children = p.ChildrenLength()
  children = [
      _deserialize_pytreedef_to_pytree(p.Children(i))
      for i in range(nr_children)
  ]
  if kind == ser_flatbuf.PyTreeDefKind.leaf:
    return 0.0
  elif kind == ser_flatbuf.PyTreeDefKind.none:
    return None
  elif kind == ser_flatbuf.PyTreeDefKind.tuple:
    return tuple(children)
  elif kind == ser_flatbuf.PyTreeDefKind.list:
    return list(children)
  elif kind == ser_flatbuf.PyTreeDefKind.dict:
    assert p.ChildrenNamesLength() == nr_children
    keys = [p.ChildrenNames(i).decode("utf-8") for i in range(nr_children)]
    return dict(zip(keys, children))
  elif kind == ser_flatbuf.PyTreeDefKind.custom:
    serialized_name = p.CustomName().decode("utf-8")
    if serialized_name not in _export.deserialization_registry:
      raise ValueError(
          "Cannot deserialize a PyTreeDef containing an "
          f"unregistered type `{serialized_name}`. "
          "Use `export.register_pytree_node_serialization` or "
          "`export.register_namedtuple_serialization`.")
    nodetype, deserialize_auxdata, from_iter = _export.deserialization_registry[serialized_name]
    auxdata = deserialize_auxdata(p.CustomAuxdataAsNumpy().tobytes())
    return from_iter(auxdata, children)
  else:
    raise ValueError(
        f"Cannot deserialize PyTreeDef with unknown kind: {kind}")


_dtype_to_dtype_kind = {
    np.dtype("bool"): ser_flatbuf.DType.bool,
    np.dtype("int8"): ser_flatbuf.DType.i8,
    np.dtype("int16"): ser_flatbuf.DType.i16,
    np.dtype("int32"): ser_flatbuf.DType.i32,
    np.dtype("int64"): ser_flatbuf.DType.i64,
    np.dtype("uint8"): ser_flatbuf.DType.ui8,
    np.dtype("uint16"): ser_flatbuf.DType.ui16,
    np.dtype("uint32"): ser_flatbuf.DType.ui32,
    np.dtype("uint64"): ser_flatbuf.DType.ui64,
    dtypes.float0: ser_flatbuf.DType.f0,
    np.dtype("float16"): ser_flatbuf.DType.f16,
    np.dtype("float32"): ser_flatbuf.DType.f32,
    np.dtype("float64"): ser_flatbuf.DType.f64,
    np.dtype("complex64"): ser_flatbuf.DType.c64,
    np.dtype("complex128"): ser_flatbuf.DType.c128,
    dtypes._bfloat16_dtype: ser_flatbuf.DType.bf16,
    dtypes._int4_dtype: ser_flatbuf.DType.i4,
    dtypes._uint4_dtype: ser_flatbuf.DType.ui4,
    dtypes._float8_e4m3b11fnuz_dtype: ser_flatbuf.DType.f8_e4m3b11fnuz,
    dtypes._float8_e4m3fn_dtype: ser_flatbuf.DType.f8_e4m3fn,
    dtypes._float8_e4m3fnuz_dtype: ser_flatbuf.DType.f8_e4m3fnuz,
    dtypes._float8_e5m2_dtype: ser_flatbuf.DType.f8_e5m2,
    dtypes._float8_e5m2fnuz_dtype: ser_flatbuf.DType.f8_e5m2fnuz,
    dtypes._float8_e3m4_dtype: ser_flatbuf.DType.f8_e3m4,
    dtypes._float8_e4m3_dtype: ser_flatbuf.DType.f8_e4m3,
    dtypes._float8_e8m0fnu_dtype: ser_flatbuf.DType.f8_e8m0fnu,
    dtypes._float4_e2m1fn_dtype: ser_flatbuf.DType.f4_e2m1fn,
}

_dtype_kind_to_dtype = {
    kind: dtype for dtype, kind in _dtype_to_dtype_kind.items()
}


def register_dtype_kind(dtype: Any, kind: int):
  _dtype_to_dtype_kind[dtype] = kind
  _dtype_kind_to_dtype[kind] = dtype


_memory_space_to_enum = {
    core.MemorySpace.Device: ser_flatbuf.MemorySpace.Device,
    core.MemorySpace.Host: ser_flatbuf.MemorySpace.Host,
    core.MemorySpace.Any: ser_flatbuf.MemorySpace.Any,
}
_memory_space_from_enum = {v: k for k, v in _memory_space_to_enum.items()}


_axis_type_to_enum = {
    core.AxisType.Auto: ser_flatbuf.AxisType.Auto,
    core.AxisType.Explicit: ser_flatbuf.AxisType.Explicit,
    core.AxisType.Manual: ser_flatbuf.AxisType.Manual,
}
_axis_type_from_enum = {v: k for k, v in _axis_type_to_enum.items()}


def _serialize_abstract_device(builder: flatbuffers.Builder,
                               device: mesh.AbstractDevice | None) -> int:
  if device is None:
    return 0
  device_kind = builder.CreateString(device.device_kind)

  ser_flatbuf.AbstractDeviceStart(builder)
  ser_flatbuf.AbstractDeviceAddDeviceKind(builder, device_kind)
  if device.num_cores is not None:
    ser_flatbuf.AbstractDeviceAddNumCores(builder, device.num_cores)
  return ser_flatbuf.AbstractDeviceEnd(builder)


def _deserialize_abstract_device(
    ser_abs_device: ser_flatbuf.AbstractDevice | None
    ) -> mesh.AbstractDevice | None:
  if ser_abs_device is None:
    return None
  device_kind = ser_abs_device.DeviceKind().decode("utf-8")
  num_cores: int | None = ser_abs_device.NumCores()
  return mesh.AbstractDevice(device_kind, num_cores)


def _serialize_abstract_mesh(builder: flatbuffers.Builder,
                             mesh: mesh.AbstractMesh) -> int:
  ser_flatbuf.AbstractMeshStartAxisSizesVector(builder, len(mesh.axis_sizes))
  for axis_size in reversed(mesh.axis_sizes):
    builder.PrependUint32(axis_size)
  axis_sizes = builder.EndVector()

  axis_names = _serialize_array(builder,
                                lambda builder, an: builder.CreateString(an),
                                mesh.axis_names)

  assert mesh.axis_types is not None, mesh
  ser_flatbuf.AbstractMeshStartAxisTypesVector(builder, len(mesh.axis_types))
  for axis_type in reversed(mesh.axis_types):
    builder.PrependByte(_axis_type_to_enum[axis_type])
  axis_types = builder.EndVector()

  abstract_device = _serialize_abstract_device(builder, mesh.abstract_device)

  ser_flatbuf.AbstractMeshStart(builder)
  ser_flatbuf.AbstractMeshAddAxisSizes(builder, axis_sizes)
  ser_flatbuf.AbstractMeshAddAxisNames(builder, axis_names)
  ser_flatbuf.AbstractMeshAddAxisTypes(builder, axis_types)
  if mesh.abstract_device is not None:
    ser_flatbuf.AbstractMeshAddAbstractDevice(builder, abstract_device)
  return ser_flatbuf.AbstractMeshEnd(builder)


def _deserialize_abstract_mesh(
  ser_mesh: ser_flatbuf.AbstractMesh) -> mesh.AbstractMesh:
  axis_sizes = tuple(ser_mesh.AxisSizes(i)
                     for i in range(ser_mesh.AxisSizesLength()))
  axis_names = tuple(ser_mesh.AxisNames(i).decode("utf-8")
                     for i in range(ser_mesh.AxisNamesLength()))
  axis_types = tuple(_axis_type_from_enum[ser_mesh.AxisTypes(i)]
                     for i in range(ser_mesh.AxisTypesLength()))
  abstract_device = _deserialize_abstract_device(ser_mesh.AbstractDevice())
  return mesh.AbstractMesh(axis_sizes, axis_names, axis_types,
                           abstract_device=abstract_device)


def _serialize_partition_spec_one_axis(builder: flatbuffers.Builder,
                                       spec: str | tuple[str, ...] | None) -> int:
  if spec is None:
    axes = ()
  else:
    axes = (spec,) if isinstance(spec, str) else spec

  axes_offset = _serialize_array(builder,
                                 lambda builder, ps: builder.CreateString(ps),
                                 axes)
  ser_flatbuf.PartitionSpecOneAxisStart(builder)
  ser_flatbuf.PartitionSpecOneAxisAddAxes(builder, axes_offset)
  return ser_flatbuf.PartitionSpecOneAxisEnd(builder)


def _deserialize_partition_spec_one_axis(
    spec: ser_flatbuf.PartitionSpecOneAxis) -> str | tuple[str, ...] | None:
  axes = tuple(spec.Axes(i).decode("utf-8") for i in range(spec.AxesLength()))
  if not axes:
    return None
  else:
    return axes[0] if len(axes) == 1 else axes


def _serialize_partition_spec(builder: flatbuffers.Builder,
                              spec: partition_spec.PartitionSpec) -> int:
  partitions = _serialize_array(builder, _serialize_partition_spec_one_axis,
                                spec._partitions)  # pyrefly: ignore[bad-argument-type]
  reduced = _serialize_array(builder,
                             lambda builder, ps: builder.CreateString(ps),
                             spec.reduced)
  unreduced = _serialize_array(builder,
                               lambda builder, ps: builder.CreateString(ps),
                               spec.unreduced)

  ser_flatbuf.PartitionSpecStart(builder)
  ser_flatbuf.PartitionSpecAddPartitions(builder, partitions)
  ser_flatbuf.PartitionSpecAddReduced(builder, reduced)
  ser_flatbuf.PartitionSpecAddUnreduced(builder, unreduced)
  return ser_flatbuf.PartitionSpecEnd(builder)


def _deserialize_partition_spec(spec: ser_flatbuf.PartitionSpec
                                ) -> partition_spec.PartitionSpec:
  partitions = tuple(_deserialize_partition_spec_one_axis(spec.Partitions(i))
                     for i in range(spec.PartitionsLength()))
  reduced = frozenset(spec.Reduced(i).decode("utf-8")
                      for i in range(spec.ReducedLength()))
  unreduced = frozenset(spec.Unreduced(i).decode("utf-8")
                        for i in range(spec.UnreducedLength()))
  return partition_spec.PartitionSpec(*partitions,
                                      reduced=reduced,
                                      unreduced=unreduced)


def _serialize_named_sharding(
    builder: flatbuffers.Builder, sharding: named_sharding.NamedSharding, *,
    uniques: _SerializedUniques
) -> int:
  abstract_mesh_idx = uniques.abstract_meshes_map[sharding.mesh.abstract_mesh]
  # TODO(necula): 1 month after 4/4/26 we can stop serializing the full
  # abstract_mesh and only serialize the index.
  mesh_offset = _serialize_abstract_mesh(builder, sharding.mesh.abstract_mesh)
  spec_offset = _serialize_partition_spec(builder, sharding.spec)
  memory_kind = builder.CreateString(sharding.memory_kind) if sharding.memory_kind is not None else 0

  ser_flatbuf.NamedShardingStart(builder)
  ser_flatbuf.NamedShardingAddMesh(builder, mesh_offset)
  ser_flatbuf.NamedShardingAddSpec(builder, spec_offset)
  if memory_kind != 0:
    ser_flatbuf.NamedShardingAddMemoryKind(builder, memory_kind)
  ser_flatbuf.NamedShardingAddAbstractMeshIdx(builder, abstract_mesh_idx)
  return ser_flatbuf.NamedShardingEnd(builder)


def _deserialize_named_sharding(
    s: ser_flatbuf.NamedSharding, *, uniques: _SerializedUniques
) -> named_sharding.NamedSharding:
  if uniques.unique_abstract_meshes:
    amesh = uniques.unique_abstract_meshes[s.AbstractMeshIdx()]
  else:
    # TODO(necula): 6 months after 4/4/26 we can stop deserializing the full
    # abstract_mesh.
    amesh = _deserialize_abstract_mesh(s.Mesh())
  spec = _deserialize_partition_spec(s.Spec())
  memory_kind = s.MemoryKind().decode("utf-8") if s.MemoryKind() is not None else None
  return named_sharding.NamedSharding(amesh, spec, memory_kind=memory_kind)


def _serialize_aval(
    builder: flatbuffers.Builder, aval: core.ShapedArray
) -> int:
  aval_kind = ser_flatbuf.AbstractValueKind.shapedArray
  shape_offsets = [builder.CreateString(str(d)) for d in aval.shape]
  ser_flatbuf.AbstractValueStartShapeVector(builder, len(aval.shape))
  for d in reversed(shape_offsets):
    builder.PrependUOffsetTRelative(d)
  shape_vector_offset = builder.EndVector()

  ser_flatbuf.AbstractValueStart(builder)
  ser_flatbuf.AbstractValueAddKind(builder, aval_kind)
  ser_flatbuf.AbstractValueAddShape(builder, shape_vector_offset)
  ser_flatbuf.AbstractValueAddDtype(builder, _dtype_to_dtype_kind[aval.dtype])
  ser_flatbuf.AbstractValueAddMemorySpace(builder, _memory_space_to_enum[aval.memory_space])
  return ser_flatbuf.AbstractValueEnd(builder)


def _deserialize_aval(aval: ser_flatbuf.AbstractValue, *,
                      scope: shape_poly.SymbolicScope,
                      sharding: named_sharding.NamedSharding | None,
                      ) -> core.ShapedArray:
  dtype = _dtype_kind_to_dtype[aval.Dtype()]
  shape = shape_poly.symbolic_shape(
      ",".join(
            aval.Shape(i).decode("utf-8") for i in range(aval.ShapeLength())
        ),
        scope=scope
    )
  if (ser_mem_space := aval.MemorySpace()):
    mem_space = _memory_space_from_enum[ser_mem_space]
  else:
    mem_space = core.MemorySpace.Device

  return core.update_aval_with_sharding(
      core.ShapedArray(shape, dtype, memory_space=mem_space), sharding
  )


def _serialize_sharding(
    builder: flatbuffers.Builder, s: _export.NamedSharding | None, *,
    uniques: _SerializedUniques) -> int:
  named_sharding = None

  if s is not None:
    named_sharding = _serialize_named_sharding(builder, s, uniques=uniques)

  ser_flatbuf.ShardingStart(builder)
  if named_sharding is not None:
    ser_flatbuf.ShardingAddNamedSharding(builder, named_sharding)
  return ser_flatbuf.ShardingEnd(builder)


def _deserialize_sharding(s: ser_flatbuf.Sharding, *,
                          uniques: _SerializedUniques) -> _export.HloSharding | named_sharding.NamedSharding | None:
  if (named_sharding_off := s.NamedSharding()) is not None:
    # After 1/15/26 all exports will have named shardings (or None)
    # TODO(necula): We must keep reading the NamedSharding for 6 months after 4/4/26
    return _deserialize_named_sharding(named_sharding_off, uniques=uniques)

  # TODO(b/489569164): We must keep reading the HloSharding for 6 months after 1/15/2026.
  if not s.HloShardingProtoIsNone():
    proto = xla_client.OpSharding()
    proto.ParseFromString(s.HloShardingProtoAsNumpy().tobytes())
    return xla_client.HloSharding.from_proto(proto)

  return None  # Unspecified sharding


def _serialize_effect(builder: flatbuffers.Builder, eff: core.Effect) -> int:
  try:
    eff_replica = eff.__class__()
  except Exception:
    raise NotImplementedError(
        f"Effect {eff} must have a nullary constructor to be serializable"
    )
  try:
    hash_eff = hash(eff)
    hash_eff_replica = hash(eff_replica)
  except Exception:
    raise NotImplementedError(
        f"Effect {eff} must be hashable to be serializable"
    )
  if eff != eff_replica or hash_eff != hash_eff_replica:
    raise NotImplementedError(
      f"Effect {eff} must have a nullary class constructor that produces an "
      "equal effect object."
    )
  effect_type_name = str(eff.__class__)
  effect_type_name_offset = builder.CreateString(effect_type_name)
  ser_flatbuf.EffectStart(builder)
  ser_flatbuf.EffectAddTypeName(builder, effect_type_name_offset)
  return ser_flatbuf.ExportedEnd(builder)


def _deserialize_effect(eff: ser_flatbuf.Effect) -> core.Effect:
  effect_type_name = eff.TypeName().decode("utf-8")
  for existing_effect_type in effects.lowerable_effects._effect_types:
    if str(existing_effect_type) == effect_type_name:
      try:
        return existing_effect_type()
      except:
        # TODO: add test
        raise NotImplementedError(
            f"deserializing effect {effect_type_name} that does not have a "
            "nullary class constructor"
        )

  raise NotImplementedError(
      f"cannot deserialize effect type {effect_type_name}"
  )


def _serialize_disabled_safety_check(
    builder: flatbuffers.Builder, check: _export.DisabledSafetyCheck
) -> int:
  custom_call_target_str = check.is_custom_call()
  custom_call_target = None
  if custom_call_target_str is not None:
    kind = ser_flatbuf.DisabledSafetyCheckKind.custom_call
    custom_call_target = builder.CreateString(custom_call_target_str)
  elif check == _export.DisabledSafetyCheck.platform():
    kind = ser_flatbuf.DisabledSafetyCheckKind.platform
  else:
    raise NotImplementedError(f"serializing DisabledSafetyCheck: {check}")

  ser_flatbuf.DisabledSafetyCheckStart(builder)
  ser_flatbuf.DisabledSafetyCheckAddKind(builder, kind)
  if custom_call_target is not None:
    ser_flatbuf.DisabledSafetyCheckAddCustomCallTarget(
        builder, custom_call_target
    )
  return ser_flatbuf.DisabledSafetyCheckEnd(builder)


def _deserialize_disabled_safety_check(
    sc: ser_flatbuf.DisabledSafetyCheck,
) -> _export.DisabledSafetyCheck:
  kind = sc.Kind()
  if kind == ser_flatbuf.DisabledSafetyCheckKind.custom_call:
    return _export.DisabledSafetyCheck.custom_call(
        sc.CustomCallTarget().decode("utf-8")
    )
  if kind == ser_flatbuf.DisabledSafetyCheckKind.platform:
    return _export.DisabledSafetyCheck.platform()

  raise ValueError(f"Cannot deserialize DisabledSafetyCheck with unknown kind: {kind}")



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import overload, Literal

from jax._src import config
from jax._src.lax import lax
from jax._src.numpy import lax_numpy as jnp
from jax._src.numpy import reductions
from jax._src.numpy import ufuncs
from jax._src.numpy.reductions import _reduction_dims, Axis
from jax._src.numpy.util import promote_args_inexact
from jax._src.typing import Array, ArrayLike

import numpy as np

# The definition of logsumexp is shared between jax.nn and jax.scipy, and
# although it matches scipy's definition, we put it here to avoid having
# unnecessary scipy dependencies.

@overload
def logsumexp(a: ArrayLike, axis: Axis = None, b: ArrayLike | None = None,
              keepdims: bool = False, return_sign: Literal[False] = False, where: ArrayLike | None = None) -> Array: ...

@overload
def logsumexp(a: ArrayLike, axis: Axis = None, b: ArrayLike | None = None,
              keepdims: bool = False, *, return_sign: Literal[True], where: ArrayLike | None = None) -> tuple[Array, Array]: ...

@overload
def logsumexp(a: ArrayLike, axis: Axis = None, b: ArrayLike | None = None,
              keepdims: bool = False, return_sign: bool = False, where: ArrayLike | None = None) -> Array | tuple[Array, Array]: ...

def logsumexp(a: ArrayLike, axis: Axis = None, b: ArrayLike | None = None,
              keepdims: bool = False, return_sign: bool = False, where: ArrayLike | None = None) -> Array | tuple[Array, Array]:
  r"""Log-sum-exp reduction.

  JAX implementation of :func:`scipy.special.logsumexp`.

  .. math::
    \operatorname{logsumexp} a = \log \sum_i b_i \exp a_i

  where the :math:`i` indices range over one or more dimensions to be reduced.

  Args:
    a: the input array
    axis: int or sequence of ints, default=None. Axis along which the sum to be
      computed. If None, the sum is computed along all the axes.
    b: scaling factors for the exponentials. Must be broadcastable to the shape of `a`.
    keepdims: If ``True``, the axes that are reduced are left in the output as
      dimensions of size 1.
    return_sign: If ``True``, the output will be a ``(result, sign)`` pair,
      where ``sign`` is the sign of the sums and ``result`` contains the
      logarithms of their absolute values. If ``False`` only ``result`` is
      returned and it will contain NaN values if the sums are negative.
    where: Elements to include in the reduction.

  Returns:
    Either an array ``result`` or a pair of arrays ``(result, sign)``, depending
    on the value of the ``return_sign`` argument.

  See also:
    :func:`jax.nn.logmeanexp`
  """
  if where is not None:
    a = jnp.where(where, a, 0)
  if b is not None:
    a_arr, b_arr = promote_args_inexact("logsumexp", a, b)
    a_arr = jnp.where(b_arr != 0, a_arr, -np.inf)
  else:
    a_arr, = promote_args_inexact("logsumexp", a)
    b_arr = a_arr  # for type checking
  pos_dims, dims = _reduction_dims(a_arr, axis)
  amax = reductions.max(a_arr.real, axis=dims, keepdims=keepdims, where=where, initial=-np.inf)
  amax = lax.stop_gradient(lax.select(ufuncs.isfinite(amax), amax, lax.full_like(amax, 0)))
  amax_with_dims = amax if keepdims else lax.expand_dims(amax, pos_dims)

  exp_a = lax.exp(lax.sub(a_arr, amax_with_dims.astype(a_arr.dtype)))
  if b is not None:
    exp_a = lax.mul(exp_a, b_arr)
  sumexp = exp_a.sum(axis=dims, keepdims=keepdims, where=where)
  sign = lax.sign(sumexp)
  if return_sign or not np.issubdtype(a_arr.dtype, np.complexfloating):
    sumexp = abs(sumexp)
  out = lax.add(lax.log(sumexp), amax.astype(sumexp.dtype))

  if return_sign:
    return (out, sign)
  if b is not None and not np.issubdtype(out.dtype, np.complexfloating):
    with config.debug_nans(False):
      out = jnp.where(sign < 0, jnp.array(np.nan, dtype=out.dtype), out)
  return out



# Copyright 2022 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
import os
import logging
from jax._src.cloud_tpu_init import running_in_cloud_tpu_vm

logger = logging.getLogger(__name__)

class ClusterEnv:
  """Interface for defining a cluster environment.

  To enable auto bootstrapping (aka :func:`jax.distributed.initialize()`),
  cluster environments need to derive from :class:`ClusterEnv` and implement
  :func:`is_env_present`, :func:`get_coordinator_address`,
  :func:`get_process_count`, and :func:`get_process_id`.
  :class:`ClusterEnv` subclasses are automatically detected when imported.
  """

  name: str = ""

  _cluster_types: list[type[ClusterEnv]] = []
  opt_in_only_method: bool = False # Override this in derived classes if necessary

  def __init_subclass__(cls, **kwargs):
    super().__init_subclass__(**kwargs)
    cls._cluster_types.append(cls)


  @classmethod
  def auto_detect_unset_distributed_params(cls,
                                           coordinator_address: str | None,
                                           num_processes: int | None,
                                           process_id: int | None,
                                           local_device_ids: Sequence[int] | None,
                                           cluster_detection_method: str | None,
                                           initialization_timeout: int | None,
                                          ) -> tuple[str | None, int | None, int | None,
                                                     Sequence[int] | None]:
    # First, we check the spec detection method because it will ignore submitted values
    # If if succeeds.
    if cluster_detection_method is not None:
      env = next( (env for env in cls._cluster_types if env.name == cluster_detection_method), None )
      if env is None:
        logger.error(f"Automatic Distributed initialization can not proceed:"
                     f" {cluster_detection_method} is not supported.")
      elif not env.is_env_present():
        logger.error(f"Automatic Distributed initialization can not proceed:"
                     f" {cluster_detection_method} is supported but not functional in this environment.")
    else:
      env = next((env for env in cls._cluster_types if env.opt_in_only_method == False and env.is_env_present()), None)

    # Above: I have wrapped the env selection in a conditional to go through
    # opt-in methods first (currently only mpi4py) but to check all possible options
    # otherwise.  Passing no cluster_detection_method results in the default, original behavior.

    if env:
      logger.debug('Initializing distributed JAX environment via %s', env.__name__)
      if coordinator_address is None:
        coordinator_port = os.environ.get("JAX_COORDINATOR_PORT")
        coordinator_address = env.get_coordinator_address(timeout_secs=initialization_timeout, override_coordinator_port=coordinator_port)
      if num_processes is None:
        num_processes = env.get_process_count()
      if process_id is None:
        process_id = env.get_process_id()
      # Never automatically set local_device_ids on TPUs
      # Defaults to single process per device if local_process_id is available.
      # This only runs if we're in a managed distributed environment.
      # Otherwise local_device_ids will remain unset,
      # which will default to all devices being visible.
      if (local_device_ids is None and not running_in_cloud_tpu_vm and
          (pid := env.get_local_process_id()) is not None):
        local_device_ids = [pid]
    else:
      logger.debug('Could not find a known environment for initializing distributed JAX. '
        'Known environments: %s', ', '.join(e.__name__ for e in cls._cluster_types))
    return (coordinator_address, num_processes, process_id, local_device_ids)

  @classmethod
  def is_env_present(cls) -> bool:
    """Returns True if process is running in this cluster environment.
    """
    raise NotImplementedError("ClusterEnv subclasses must implement is_env_present")

  @classmethod
  def get_coordinator_address(cls, timeout_secs: int | None, override_coordinator_port: str | None) -> str:
    """Returns address and port used by JAX to bootstrap.

    Process id 0 will open a tcp socket at "hostname:port" where
    all the processes will connect to initialize the distributed JAX service.
    The selected port needs to be free.
    :func:`get_coordinator_address` needs to return the same hostname and port on all the processes.

    Returns:
      "hostname:port"
    """
    raise NotImplementedError("ClusterEnv subclasses must implement get_coordinator_address")

  @classmethod
  def get_process_count(cls) -> int:
    raise NotImplementedError("ClusterEnv subclasses must implement get_process_count")

  @classmethod
  def get_process_id(cls) -> int:
    raise NotImplementedError("ClusterEnv subclasses must implement get_process_id")

  @classmethod
  def get_local_process_id(cls) -> int | None:
    """ Get index of current process inside a host.

    The method is only useful to support single device per process.
    In that case, each process will see a local device whose ID is
    the same as its local process ID.
    If None, JAX will not restrict the visible devices.
    """
    return None



# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Layout utilities."""

from typing import assert_never

from jax._src.lib import mosaic_gpu_dialect as mgpu
from jax._src.lib.mlir import ir

from . import fragmented_array as fa
from . import launch_context


# TODO(b/415721295): Refine return type once minimum jaxlib version is 0.10.0
def _to_splat_fragmented_layout_attr(
    layout: fa.WGSplatFragLayout,
) -> ir.Attribute:
  """Constructs a #mosaic_gpu.WGSplatFragLayout attribute from a WGSplatFragLayout."""
  shape = ir.DenseI64ArrayAttr.get(layout.shape)
  return mgpu.WGSplatFragLayoutAttr.get(shape)


def _from_splat_fragmented_layout_attr(
    attr: ir.Attribute,
) -> fa.WGSplatFragLayout:
  # TODO(b/415721295): Refine arg type once minimum jaxlib version is 0.10.0
  assert isinstance(attr, mgpu.WGSplatFragLayoutAttr)
  return fa.WGSplatFragLayout(shape=tuple(attr.shape))


# TODO(b/415721295): Refine return type once minimum jaxlib version is 0.10.0
def _to_strided_fragmented_layout_attr(
    layout: fa.WGStridedFragLayout,
) -> ir.Attribute:
  """Constructs a #mosaic_gpu.WGStridedFragLayout attribute from a WGStridedFragLayout."""
  shape = ir.DenseI64ArrayAttr.get(layout.shape)
  return mgpu.WGStridedFragLayoutAttr.get(shape, layout.vec_size)


def _from_strided_fragmented_layout_attr(
    attr: ir.Attribute,
) -> fa.WGStridedFragLayout:
  """Constructs a WGStridedFragLayout from a #mosaic_gpu.WGStridedFragLayout attribute."""
  # TODO(b/415721295): Refine arg type once minimum jaxlib version is 0.10.0
  assert isinstance(attr, mgpu.WGStridedFragLayoutAttr)
  return fa.WGStridedFragLayout(
      shape=tuple(attr.shape),
      vec_size=attr.vector_size,
  )


# TODO(b/415721295): Refine return type once minimum jaxlib version is 0.10.0
def _to_tiled_layout_attr(
    layout: fa.TiledLayout,
) -> ir.Attribute:
  """Constructs a #mosaic_gpu.TiledLayout attribute from a TiledLayout."""
  i64 = ir.IntegerType.get_signless(64)

  def _int_or_replicated(d: int | fa.Replicated) -> ir.Attribute:
    if isinstance(d, fa.Replicated):
      return mgpu.ReplicatedAttr.get(d.times)
    return ir.IntegerAttr.get(i64, d)

  def _tile_attr(tile):
    return ir.ArrayAttr.get([ir.IntegerAttr.get(i64, d) for d in tile])

  tiling_attr = ir.ArrayAttr.get(
      [_tile_attr(tile) for tile in layout.tiling.tiles]
  )
  warp_dims_attr = ir.ArrayAttr.get(
      [_int_or_replicated(d) for d in layout.warp_dims]
  )
  lane_dims_attr = ir.ArrayAttr.get(
      [_int_or_replicated(d) for d in layout.lane_dims]
  )

  return mgpu.TiledLayoutAttr.get(
      tiling_attr, warp_dims_attr, lane_dims_attr, layout.vector_dim
  )


def _from_tiled_layout_attr(
    attr: ir.Attribute,
) -> fa.TiledLayout:
  """Constructs a TiledLayout from a #mosaic_gpu.TiledLayout attribute."""
  # TODO(allanrenucci): Refine arg type once minimum jaxlib version is 0.10.0
  assert isinstance(attr, mgpu.TiledLayoutAttr)

  def _from_int_or_replicated_attr(d_attr: ir.Attribute) -> int | fa.Replicated:
    if isinstance(d_attr, mgpu.ReplicatedAttr):
      return fa.Replicated(times=mgpu.ReplicatedAttr(d_attr).times)
    return ir.IntegerAttr(d_attr).value

  tiles = tuple(
      tuple(ir.IntegerAttr(d).value for d in ir.ArrayAttr(tile))
      for tile in attr.tiling
  )
  warp_dims = tuple(_from_int_or_replicated_attr(d) for d in attr.warp_dims)
  lane_dims = tuple(_from_int_or_replicated_attr(d) for d in attr.lane_dims)

  return fa.TiledLayout(
      tiling=fa.Tiling(tiles),
      warp_dims=warp_dims,
      lane_dims=lane_dims,
      vector_dim=attr.vector_dim,
  )


def to_layout_attr(layout: fa.FragmentedLayout) -> ir.Attribute:
  """Constructs an MLIR attribute that corresponds to the given layout."""
  match layout:
    case fa.WGSplatFragLayout():
      return _to_splat_fragmented_layout_attr(layout)
    case fa.WGStridedFragLayout():
      return _to_strided_fragmented_layout_attr(layout)
    case fa.TiledLayout():
      return _to_tiled_layout_attr(layout)
    case _:
      assert_never(layout)


def from_layout_attr(attr: ir.Attribute) -> fa.FragmentedLayout:
  """Constructs a layout from an MLIR attribute."""
  if isinstance(attr, mgpu.WGSplatFragLayoutAttr):
    return _from_splat_fragmented_layout_attr(attr)
  elif isinstance(attr, mgpu.WGStridedFragLayoutAttr):
    return _from_strided_fragmented_layout_attr(attr)
  elif isinstance(attr, mgpu.TiledLayoutAttr):
    return _from_tiled_layout_attr(attr)
  else:
    raise NotImplementedError(
        f"Unsupported layout for conversion from MLIR attribute: {attr}"
    )


def splat_is_compatible_with_tiled(
    l1: fa.WGSplatFragLayout, l2: fa.TiledLayout
) -> bool:
  # A splat layout is compatible with a tiled layout up to replication if each
  # dimension in the shape of the splat layout is divisible by the corresponding
  # dimension in the base tile shape.
  s1, s2 = l1.shape, l2.base_tile_shape
  return all(d1 % d2 == 0 for d1, d2 in zip(s1, s2))


def to_transform_attr(
    transform: launch_context.MemRefTransform | mgpu.SwizzlingMode,
) -> ir.Attribute:
  if isinstance(transform, launch_context.TileTransform):
    return mgpu.TileTransformAttr.get(transform.tiling)
  elif isinstance(transform, launch_context.TransposeTransform):
    return mgpu.TransposeTransformAttr.get(transform.permutation)
  elif isinstance(transform, mgpu.SwizzlingMode):
    return mgpu.SwizzleTransformAttr.get(transform)
  else:
    raise NotImplementedError(f"Unsupported transform {transform}")


def from_transform_attr(
    transform: ir.Attribute,
) -> launch_context.MemRefTransform | mgpu.SwizzlingMode:
  if isinstance(transform, mgpu.TileTransformAttr):
    return launch_context.TileTransform(
        tuple(mgpu.TileTransformAttr(transform).tiling)
    )
  elif isinstance(transform, mgpu.TransposeTransformAttr):
    return launch_context.TransposeTransform(
        tuple(mgpu.TransposeTransformAttr(transform).permutation)
    )
  elif isinstance(transform, mgpu.SwizzleTransformAttr):
    return mgpu.SwizzlingMode(mgpu.SwizzleTransformAttr(transform).swizzle)
  else:
    raise NotImplementedError(f"Unsupported transform {transform}")
